# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# widget.py - base widgets class
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Widget' ]

from gi.repository import Clutter as clutter

class Widget(object):

    # clutter object
    obj = None

    # internal id of the widget on client and server
    wid = None

    ALIGN_LEFT = 'left'
    ALIGN_RIGHT = 'right'
    ALIGN_TOP = 'top'
    ALIGN_BOTTOM = 'bottom'
    ALIGN_CENTER = 'center'
    ALIGN_SHRINK = 'shrink'

    def prepare(self, modified):
        """
        Prepare rendering
        Executed in the kaa mainloop
        """
        pass

    def send_widget_event(self, *args):
        """
        Send an event to the candy widget
        """
        self.server.send_event('widget_call', self.wid, *args)
        
    def create(self):
        """
        Create the clutter object
        Executed in the clutter thread
        """
        pass

    def delete(self):
        """
        Delete the clutter object
        Executed in the clutter thread
        """
        if self.obj and self.obj.get_parent():
            self.obj.get_parent().remove_actor(self.obj)
        self.obj = None

    def reparent(self, parent, sibling):
        """
        Reparent the clutter object
        Executed in the clutter thread
        """
        if self.obj.get_parent():
            self.obj.get_parent().remove_actor(self.obj)
        if parent and sibling and sibling.obj.get_parent():
            parent.obj.insert_child_above(self.obj, sibling.obj)
        elif parent and sibling:
            parent.obj.add_actor(self.obj)
        elif parent:
            parent.obj.add_actor(self.obj)

    def set_position(self):
        """
        Set a new position without re-rendering the widget
        Executed in the clutter thread
        """
        self.obj.set_position(self.x, self.y)

    def raise_top(self):
        """
        Raise widget to the top of the stack
        """
        self.obj.raise_top()

    def lower_bottom(self):
        """
        Lower widget to the bottom of the stack
        """
        self.obj.lower_bottom()

    def above_sibling(self, sibling):
        """
        Move the widget above the given sibling
        """
        self.obj.get_parent().set_child_above_sibling(self.obj, sibling.obj)

    def update(self, modified):
        """
        Render the widget
        Executed in the clutter thread
        """
        if 'width' in modified and self.width:
            self.obj.set_width(self.width)
        if 'height' in modified and self.height:
            self.obj.set_height(self.height)
        if 'opacity' in modified:
            self.obj.set_opacity(self.opacity)
        if 'scale_x' in modified or 'scale_y' in modified:
            self.obj.set_scale(self.scale_x, self.scale_y)
        if 'anchor_point' in modified:
            self.obj.set_anchor_point(*self.anchor_point)
        if 'visible' in modified:
            if self.visible:
                self.obj.show()
            else:
                self.obj.hide()

    def animate(self, ease, secs, *args):
        # Note: it is not possible to stop or modify an animation
        # right now. The result from this call has to go back to the
        # main app somehow.
        keys = []
        values = []
        args = list(args)
        while args:
            keys.append(args.pop(0))
            values.append(args.pop(0))
        self.obj.animatev(getattr(clutter.AnimationMode, ease), int(secs * 1000) or 1, keys, values)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# rectangle.py - rectangle widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Rectangle' ]

from widget import Widget
from ..core import Color

class Rectangle(Widget):
    candyxml_name = 'rectangle'
    candy_backend = 'candy.Rectangle'
    attributes = [ 'color', 'border_size', 'border_color', 'radius' ]

    attribute_types = {
        'color': Color,
        'border_color': Color
    }

    def __init__(self, pos=None, size=None, color=None, border_size=0,
                 border_color=None, radius=0, context=None):
        """
        Create a Rectangle widget

        @param pos: (x,y) position of the widget or None
        @param size: (width,height) geometry of the widget.
        @param color: kaa.candy.Color to fill the rectangle
        @param border_size: size of the rectangle border
        @param border_color: kaa.candy.Color of the border. This argument is
            not needed when border_size is 0.
        @param radius: radius for a rectangle with round edges.
        """
        super(Rectangle, self).__init__(pos, size, context)
        self.color = color
        self.border_size = border_size
        self.border_color = border_color
        self.radius = radius

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
            <rectangle x='0' y='0' width='100' height='100' color='0xff0000'
                border_size='2' border_color='0x000000' radius='10'/>
        """
        return super(Rectangle, cls).candyxml_parse(element).update(
            radius=int(element.radius or 0), border_size=int(element.border_size or 0),
            color=element.color, border_color=element.border_color)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# __init__.py - kaa.candy main entry point
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

# python imports
import sys
import os

if not sys.argv[0].startswith(os.path.dirname(__file__)):
    # only import the submodules from the main application and not
    # from the rendering backend process.
    from core import Context, Color, Font
    from modifier import Modifier, Properties, Eventhandler, Dependency
    from template import is_template, Template
    from widgets import *
    from stage import Stage, Layer

    import candyxml
    import config

    ALIGN_LEFT = Widget.ALIGN_LEFT
    ALIGN_CENTER = Widget.ALIGN_CENTER
    ALIGN_RIGHT = Widget.ALIGN_RIGHT
    ALIGN_TOP = Widget.ALIGN_TOP
    ALIGN_BOTTOM = Widget.ALIGN_BOTTOM
    ALIGN_SHRINK = Widget.ALIGN_SHRINK



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# widgets - backend files for the widgets
# -----------------------------------------------------------------------------
# This directory contains the clutter code executed in the rendering
# process and not the process importing kaa.candy.
#
# -----------------------------------------------------------------------------
# kaa-candy - Third generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer, Jason Tackaberry
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = []

import logging
import kaa.utils

# get logging object
log = logging.getLogger('candy')

for name, module in kaa.utils.get_plugins(location=__file__).items():
    if not hasattr(module, '__all__'):
        # This should never happen and is for developing inside
        # kaa.candy only. Still, we should use the logging module
        # somehow and get the logging info to the main process.
        log.error('unable to import %s: %s' % (name, module))
        continue
    for widget in module.__all__:
        __all__.append(widget)
        globals()[widget] = getattr(module, widget)




# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# label.py - text label widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Label' ]

# python
import re
import kaa.base

# kaa.candy imports
from widget import Widget
from ..core import Color, Font

class Label(Widget):
    """
    Text label widget
    """
    candyxml_name = 'label'
    candy_backend = 'candy.Label'
    attributes = [ 'color', 'font', 'text' ]
    attribute_types = {
        'color': Color,
        'font': Font
    }

    __text_regexp = re.compile('\$([a-zA-Z][a-zA-Z0-9_\.]*)|\${([^}]*)}')
    __text = None

    def __init__(self, pos=None, size=None, color=None, font=None, text='', context=None):
        """
        Create a new label widget

        @param pos: (x,y) position of the widget or None
        @param text: Text to render. This can also be a context based string like
            C{$text} with the context containing C{text='My String'}. This will
            make the widget context sensitive.
        @param size: (width,height) geometry of the widget.
        @param font: kaa.candy.Font object
        @param context: the context the widget is created in
        """
        super(Label, self).__init__(pos, size, context)
        self.color = color
        self.font = font
        self.text = text

    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, text):
        self.__text_provided = text
        def replace_context(matchobj):
            match = matchobj.groups()[0] or matchobj.groups()[1]
            s = self.context.get(match, '')
            if s is not None:
                return kaa.base.py3_str(s, coerce=True)
            return ''
        if self.context:
            # we have a context, use it
            text = re.sub(self.__text_regexp, replace_context, text)
        if self.__text == text:
            return
        self.__text = text
        self.queue_rendering()

    def sync_context(self):
        """
        Adjust to a new context
        """
        self.text = self.__text_provided

    def sync_layout(self, size):
        """
        Sync layout changes and calculate intrinsic size based on the
        parent's size.
        """
        super(Label, self).sync_layout(size)
        if self.font.size == 0:
            # FIXME: wrong when the widget size changes
            self.font = self.font.get_font(self.height)
        width, height = self.font.get_width(self.text), self.font.get_height(Font.MAX_HEIGHT)
        if self.width and width > self.width and self.width > 0:
            width = self.width
        self.intrinsic_size = width, height

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
          <label y='50' width='100' font='Vera:24' color='0xffffffff'>
              text to show
          </label>
        The text can also be a context based variable like C{$text}. This
        will make the widget context sensitive.
        """
        return super(Label, cls).candyxml_parse(element).update(
            font=element.font, color=element.color, text=element.content)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# template.py - Template for creating a widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'is_template', 'Template' ]

# python imports
import logging
import traceback

# kaa.candy imports
from core import Context
from modifier import Modifier
from candyxml import get_class

# get logging object
log = logging.getLogger('kaa.candy')


def is_template(obj):
    """
    Returns True if the given object is a kaa.candy template class. This function
    is needed to check if a given widget is a real clutter widget or only a template
    for creating one.
    """
    return isinstance(obj, Template)


class Template(object):
    """
    Template to create a widget on demand. All XML parsers will create such an
    object to parse everything at once.
    """

    def __init__(self, cls, name, **kwargs):
        """
        Create a template for the given class

        :param cls: widget class
        :param kwargs: keyword arguments for cls.__init__
        """
        self._cls = cls
        self._modifier = kwargs.pop('modifier', [])
        self._kwargs = kwargs
        self.name = name

    def __call__(self, context=None, **kwargs):
        """
        Create the widget with the given context and override some
        constructor arguments.

        :param context: context to create the widget in
        :returns: widget object
        """
        if context is not None:
            context = Context(context)
        args = self._kwargs.copy()
        args.update(kwargs)
        args['context'] = context
        try:
            widget = self._cls(**args)
            widget.__template__ = self
            widget.name = self.name
        except Exception, e:
            log.exception('unable to create widget %s', self._cls)
            raise RuntimeError('unable to create %s%s: %s' % (self._cls, args.keys(), e))
        for modifier in self._modifier:
            widget = modifier.modify(widget)
        return widget

    def __repr__(self):
        return '<kaa.candy.Template for %s>' % self._cls

    @classmethod
    def candyxml_get_class(cls, element):
        """
        Get the class for the candyxml element. This function may be overwritten
        by inheriting classes and should not be called from outside such a class.
        """
        return get_class(element.node, element.style)

    @classmethod
    def candyxml_create(cls, element):
        """
        Parse the candyxml element for parameter and create a Template.
        """
        modifier = []
        for subelement in element.get_children():
            mod = Modifier.candyxml_create(subelement)
            if mod is not None:
                modifier.append(mod)
                element.remove(subelement)
        widget = cls.candyxml_get_class(element)
        if widget is None:
            log.error('undefined widget %s:%s', element.node, element.style)
        kwargs = widget.candyxml_parse(element)
        if modifier:
            kwargs['modifier'] = modifier
        template = cls(widget, name=kwargs.candyname, **kwargs)
        return template



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# modifier.py - Modifier for the XML Parser and Templates
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Modifier', 'Properties', 'Eventhandler', 'Dependency' ]

# python imports
import logging

# kaa imports
import kaa

# get logging object
log = logging.getLogger('kaa.candy')

class Modifier(object):
    """
    Modifier base class for classes that change widgets on creation by
    templates. In the XML file they are added as subnode to the widget
    to change. Examples are Properties and ReflectionModifier.
    """

    class __metaclass__(type):
        def __new__(meta, name, bases, attrs):
            cls = type.__new__(meta, name, bases, attrs)
            if 'candyxml_name' in attrs.keys():
                if cls.candyxml_name in Modifier._candyxml_modifier:
                    raise RuntimeError('%s already defined' % cls.candyxml_name)
                Modifier._candyxml_modifier[cls.candyxml_name] = cls
            return cls

    _candyxml_modifier = {}

    def modify(self, widget):
        """
        Modify the given widget.
        @param widget: widget to modify
        @returns: changed widget (may be the same)
        """
        raise NotImplementedError

    @classmethod
    def candyxml_create(cls, element):
        """
        Create the modifier for the given element.

        @note: do not call this function from inheriting functions. The name
            is the same but the logic is different. This functions calls the
            implementation variant, not the other way around.
        """
        cls = Modifier._candyxml_modifier.get(element.node)
        if cls is None:
            return cls
        return cls.candyxml_create(element)


class Properties(dict, Modifier):
    """
    Properties class to apply the given properties to a widget. This is a
    dictionary for clutter functions to call after the widget is created.
    """

    #: candyxml name
    candyxml_name = 'properties'

    def modify(self, widget):
        """
        Apply to the given widget.

        @param widget: a kaa.candy.Widget
        """
        for key, value in self.items():
            setattr(widget, key.replace('-', '_'), value)
        return widget

    @classmethod
    def candyxml_create(cls, element):
        """
        Parse the candyxml element and create a Properties object::

          <widget>
            <properties key=value key=value/>
          </widget>
        """
        properties = cls()
        for key, value in element.attributes():
            if key in ('opacity', 'depth', 'scale_x', 'scale_y'):
                value = int(value)
            elif key in ('rotation','xrotation','yrotation','zrotation'):
                value = float(value)
            elif key in ('xalign', 'yalign'):
                value = value.lower()
            elif key in ('keep_aspect',):
                value = value.lower() in ('yes', 'true')
            elif key in ('anchor_point', ):
                value = [ int(x) for x in value.split(',') ]
            properties[key] = value
        return properties


class Eventhandler(dict, Modifier):
    """
    Eventhandler class to apply the given events to a widget.
    """

    #: candyxml name
    candyxml_name = 'event'
    signatures = {
        'replace': 'prev, next',
        'create': 'widget',
    }
    def modify(self, widget):
        """
        Apply to the given widget.

        @param widget: a kaa.candy.Widget
        """
        widget._candy_events.update(self)
        return widget

    @classmethod
    def candyxml_create(cls, element):
        """
        Parse the candyxml element and create an Eventhandler object::

          <widget>
            <event name=value script=value/>
          </widget>
        """
        obj = cls()
        for event in ('%s-%s' % (element.parent.node, element.name), element.name):
            signature = cls.signatures.get(event, None)
            if signature:
                break
        else:
            raise RuntimeError('unable to parse event %s-%s' % (element.parent.node, element.name))
        if element.content.strip():
            func = 'def func(%s):\n' % signature
            if element.content.find('yield') >= 0:
                func = '@kaa.coroutine()\n' + func
            if element.content.strip().find('\n') == -1:
                func += '    '
            func += element.content.rstrip().lstrip('\n') + '\n'
            try:
                exec(func)
                obj[event] = func
                return obj
            except Exception, e:
                log.exception('unable to parse event')
        scripts = element.candyxml.scripts
        obj[event] = scripts[element.script]
        return obj


class Dependency(list, Modifier):
    """
    Dependency class to add the given dependencies to a widget forcing
    a replace when the context changes.
    """

    #: candyxml name
    candyxml_name = 'replace-on-context-change'

    eventhandler = None

    def modify(self, widget):
        """
        Apply to the given widget.

        @param widget: a kaa.candy.Widget
        """
        widget.add_dependencies(*self)
        if self.eventhandler:
            return self.eventhandler.modify(widget)
        return widget

    @classmethod
    def candyxml_create(cls, element):
        """
        Parse the candyxml element and create a Dependency object::

          <widget>
            <replace-on-context-change key="var1 var2"/>
          </widget>
        """
        obj = cls()
        dependency=(element.key or element.keys or '').strip().split(' ')
        while '' in dependency:
            dependency.remove('')
        obj.extend(dependency)
        if element.script or element.content:
            element.name = 'replace'
            obj.eventhandler = Eventhandler.candyxml_create(element)
        return obj



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# group.py - group widget containing other widgets
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Group' ]

from gi.repository import Clutter as clutter

import widget

class Group(widget.Widget):

    def create(self):
        """
        Create the clutter object
        """
        self.obj = clutter.Group.new()
        self.obj.show()

    def hide(self):
        """
        Hide the widget
        """
        self.obj.hide()

    def show(self):
        """
        Show the widget
        """
        self.obj.show()

    def update(self, modified):
        """
        Render the widget
        """
        super(Group, self).update(modified)
        if 'clip' in modified:
            if self.clip:
                (x, y), (width, height) = self.clip
                self.obj.set_clip(x, y, width, height)
            else:
                self.obj.remove_clip()



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# thumbnail.py - thumbnail widget for kaa.beacon thumbnail objects
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2013 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Thumbnail' ]

# kaa imports
import kaa.beacon

# kaa.candy imports
from widget import Widget
from image import Image

class Thumbnail(Image):
    """
    Widget showing a kaa.beacon.Thumbnail
    """
    candyxml_name = 'thumbnail'
    keep_aspect = True

    def __init__(self, pos, size, thumbnail=None, default=None, context=None):
        """
        Create the Thumbnail widget

        @param pos: (x,y) position of the widget or None
        @param size: (width,height) geometry of the widget
        @param thumbnail: kaa.beacon.Thumbnail object or a string which points
            to the Thumbnail object in the context.
        @param context: the context the widget is created in
        """
        super(Thumbnail, self).__init__(pos, size, context=context)
        self.set_thumbnail(thumbnail, default)

    def set_thumbnail(self, thumbnail, default=None):
        """
        Set the thumbnail and a default if the thumbnail is invalid
        """
        self.__thumbnail_provided = thumbnail
        self.__default_provided = default
        if isinstance(thumbnail, (str, unicode)):
            # get thumbnail from context
            # FIXME: make this dynamic
            thumbnail = self.context.get(thumbnail)
        item = None
        if isinstance(thumbnail, kaa.beacon.Item):
            item, thumbnail = thumbnail, thumbnail.get('thumbnail')
        self._thumbnail = thumbnail
        if default and not default.startswith('/'):
            default = self.search(default)
        if default:
            self.image = default
        if self._thumbnail is not None:
            # show thumbnail
            self._beacon_thumbnail_ready(force=True)
        elif item is not None and not item.scanned:
            scanning = item.scan()
            if scanning:
                scanning.connect_weak_once(self._beacon_update, item)

    def sync_context(self):
        """
        Adjust to a new context
        """
        self.set_thumbnail(self.__thumbnail_provided, self.__default_provided)

    def _beacon_update(self, changes, item):
        self._thumbnail = item.get('thumbnail')
        if self._thumbnail is not None:
            return self._beacon_thumbnail_ready(force=True)

    def _beacon_thumbnail_ready(self, force=False):
        """
        Callback to render the thumbnail to the texture.
        @todo: add thumbnail update based on beacon mtime
        @todo: try to force large thumbnails
        """
        image = self._thumbnail.image
        if image:
            self.image = image
        elif self._thumbnail.failed:
            return False
        if force and self._thumbnail.needs_update:
            # Create thumbnail; This object will hold a reference to the
            # beacon.Thumbnail object and uses high priority. Since we
            # only connect weak to the result we loose the thumbnail object
            # when this widget is removed which will reduce the priority
            # to low. This is exactly what we want. The create will be
            # scheduled in the mainloop but since we do not wait it is ok.
            self._thumbnail.create(self._thumbnail.PRIORITY_HIGH).\
                connect_weak_once(self._beacon_thumbnail_ready)

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
          <thumbnail x='10' y='10' width='100' height='100' thumbnail='thumb'/>
        The thumbnail parameter must be a string that will be evaluated based
        on the given context.
        """
        return Widget.candyxml_parse(element).update(
            thumbnail=element.thumbnail, default=element.default)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# image.py - image widget based on kaa.imlib2
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2013 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Image' ]

# python imports
import os
import logging
import hashlib
import struct
import tempfile

# kaa imports
import kaa
import kaa.net.url
import kaa.imlib2

# kaa.candy imports
from widget import Widget
from .. import config

# get logging object
log = logging.getLogger('kaa.candy')

class Image(Widget):
    """
    Image widget
    """
    candyxml_name = 'image'
    candy_backend = 'candy.ImageTexture'

    attributes = [ 'sync_data', 'modified', 'keep_aspect', 'load_async' ]

    # image variables
    modified = True
    keep_aspect = False
    load_async = False

    # class variable with a dict of images currently loading
    __downloads = {}

    __filename = None
    __image = None

    class _Info(object):
        def __init__(self, data, width, height):
            self.data = data
            self.width = width
            self.height = height

    def __init__(self, pos=None, size=None, url=None, context=None):
        """
        Create the Image

        @param pos: (x,y) position of the widget or None
        @param size: (width,height) geometry of the widget
        @param url: filename or url of the image. If the image is not found
            config.imagepath will be searched for the image. If the url
            starts with C{$} the url will be searched in the context.
        @param context: the context the widget is created in
        """
        super(Image, self).__init__(pos, size, context)
        self.image = url

    def sync_context(self):
        """
        Adjust to a new context
        """
        self.image = self.__image_provided

    def sync_layout(self, size):
        """
        Sync layout changes and calculate intrinsic size based on the
        parent's size.
        """
        if not self.__image:
            self.intrinsic_size = 0, 0
            return 0, 0
        super(Image, self).sync_layout(size)
        width, height = self.size
        if self.keep_aspect:
            aspect = float(self.__image.width) / self.__image.height
            if int(height * aspect) > width:
                height = int(width / aspect)
            else:
                width = int(height * aspect)
            self.intrinsic_size = width, height

    def sync_prepare(self):
        """
        Prepare widget for the next sync with the backend
        """
        if not self.modified:
            return False
        if not self.__image:
            self.sync_data = None
        elif isinstance(self.__image.data, (str, unicode)):
            self.sync_data = self.__image.data, False
        else:
            fd, filename = tempfile.mkstemp(prefix='candy', suffix='.png', dir='/dev/shm')
            os.close(fd)
            try:
                # write the image data to shm
                self.__image.data.save(filename)
                self.sync_data = filename, True
            except Exception, e:
                log.error('unable to save imlib2 image')
                self.sync_data = None
                self.modified = False
        self.modified = False
        return True

    def search(self, name):
        """
        Helper function to get the full path of the image.
        """
        for path in config.imagepath:
            filename = os.path.join(path, name)
            if os.path.isfile(filename):
                return filename
        return None

    def get_cachefile(self, url):
        """
        Return the cache filename for the given url
        """
        base = hashlib.md5(url).hexdigest() + os.path.splitext(url)[1]
        return kaa.tempfile('candy-images/' + base)

    def _download_complete(self, status, cachefile):
        """
        Callback for HTTP GET result. The image should be in the
        cachefile.
        """
        if cachefile in self.__downloads:
            del self.__downloads[cachefile]
        self.image = cachefile

    @property
    def image(self):
        """
        Return the image info object
        """
        return self.__image

    @image.setter
    def image(self, image):
        """
        Set a new image. Either a kaa.imlib2.Image or based on a
        filename or url.
        """
        self.__image_provided = image
        if image and isinstance(image, (str, unicode)) and image.startswith('$'):
            # variable from the context, e.g. $varname
            image = self.context.get(image) or ''
        if isinstance(image, kaa.imlib2.Image):
            # provided image is an imlib2 image object
            self.__image = Image._Info(image, image.width, image.height)
            self.modified = True
            return
        if image and image.startswith('http://'):
            # remote image, create local cachefile
            # FIXME: how to handle updates on the remote side?
            cachefile = self.get_cachefile(image)
            if not os.path.isfile(cachefile):
                # Download the image
                # FIXME: errors will be dropped
                # FIXME: support other remote sources
                # FIXME: use one thread (jobserver) for all downloads
                #  or at least a max number of threads to make the individual
                #  image loading faster
                if not cachefile in self.__downloads:
                    tmpfile = cachefile + '.tmp'
                    c = kaa.net.url.fetch(image, cachefile, tmpfile)
                    self.__downloads[cachefile] = c
                self.__downloads[cachefile].connect_weak_once(self._download_complete, cachefile)
                image = None
            else:
                image = cachefile
        if image and not image.startswith('/'):
            # try to locate the image in our image path
            image = self.search(image)
        if self.__filename == image:
            # unchanged filename
            return
        self.__filename = image
        if not image:
            # image is not valid
            log.error('invalid image: %s', self.__image_provided)
            self.__image = None
            self.modified = True
            return
        iheight = iwidth = 0
        try:
            # try to get the image geometry
            fd = open(image)
            if fd.read(2) == '\xff\xd8':
                # image looks like a JPG file. Let us see if we can
                # get the geometry from the metadata and then use the
                # clutter backend to actually load the image
                app = fd.read(4)
                while (len(app) == 4):
                    (ff,segtype,seglen) = struct.unpack(">BBH", app)
                    if ff != 0xff or segtype == 0xd9:
                        break
                    if segtype >= 0xC0 and segtype <= 0xCF:
                        iheight, iwidth = struct.unpack('>BHHB', fd.read(seglen-2)[:6])[1:3]
                        break
                    fd.seek(seglen-2,1)
                    app = fd.read(4)
            fd.seek(0)
            if fd.read(8) == '\211PNG\r\n\032\n':
                # image looks like a PNG file. Let us see if we can
                # get the geometry from the metadata and then use the
                # clutter backend to actually load the image
                while True:
                    (seglen, segtype) = struct.unpack('>I4s', fd.read(8))
                    if segtype == 'IEND':
                        break
                    if segtype == 'IHDR':
                        data = fd.read(seglen+4)
                        iwidth, iheight = struct.unpack(">IIb", data[:9])[:2]
                        break
                    fd.seek(seglen+4,1)
            if iheight > 0 and iwidth > 0:
                # the backend can load the image itself
                self.__image = Image._Info(image, iwidth, iheight)
            else:
                # load using imlib2; the backend may not be able to
                # load the image itself
                image = kaa.imlib2.Image(image)
                self.__image = Image._Info(image, image.width, image.height)
        except Exception, e:
            log.error('unable to load %s', image)
            self.__image = None
        finally:
            try:
                fd.close()
            except:
                # if open failed, close will also fail
                pass
        self.modified = True

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. Example::
          <image x='10' y='10' width='200' height='100' filename='image.jpg'/>
          <image width='200' height='100' url='http://www.exapmple.com/image.jpg'/>
        """
        return super(Image, cls).candyxml_parse(element).update(
            url=element.url or element.filename)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# widgets - backend files for the widgets
# -----------------------------------------------------------------------------
# This directory contains the video and audio player widgets. This
# module should not be accessed directly and exposes all its players
# to candy.players
#
# -----------------------------------------------------------------------------
# kaa-candy - Third generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer, Jason Tackaberry
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = []

import logging
import kaa.utils
import candy

# get logging object
log = logging.getLogger('candy')

for name, module in kaa.utils.get_plugins(location=__file__).items():
    if not hasattr(module, 'Player'):
        # This should never happen and is for developing inside
        # kaa.candy only. Still, we should use the logging module
        # somehow and get the logging info to the main process.
        log.error('disable player %s: %s' % (name, module))
        continue
    candy.player[name] = getattr(module, 'Player')

def init(server):
    """
    Called after import
    """
    server.send_event('player', candy.player.keys())



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# text.py - text widget
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2013 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Text' ]

from gi.repository import Pango, PangoCairo

import image

class Text(image.CairoTexture):

    def draw(self, cr):
        """
        Render the cairo context
        """
        if self.color:
            cr.set_source_rgba(*self.color.to_cairo())
        layout = PangoCairo.create_layout(cr)
        layout.set_width(self.width * Pango.SCALE)
        layout.set_height(self.height * Pango.SCALE)
        layout.set_ellipsize(Pango.EllipsizeMode.END)
        layout.set_alignment(getattr(Pango.Alignment, str(self.align).upper()))
        layout.set_wrap(Pango.WrapMode.WORD_CHAR)
        layout.set_font_description(self.font.get_font_description())
        layout.set_text(self.text, -1)
        PangoCairo.show_layout(cr, layout)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# gstreamer.py - Gstreamer based player
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

# Python imports
import os
import sys
import logging
import gi

# get logging object
log = logging.getLogger('candy')

gi.require_version('Gst', '1.0')

# Clutter and GStreamer GI bindings
from gi.repository import Clutter as clutter, Gst as gst, GObject as gobject

gst.init([])

GST_PLAY_FLAG_TEXT = (1 << 2)

import kaa.metadata
import candy

def requires_state(*states):
    """
    Small internal state maschine for gstreamer calls from the
    frontend.
    """
    def decorator(func):
        def newfunc(self, *args, **kwargs):
            if not self.state in states:
                return self.delayed.append((states, func, args, kwargs))
            return func(self, *args, **kwargs)
        return newfunc
    return decorator

SEEK_RELATIVE = 'SEEK_RELATIVE'
SEEK_ABSOLUTE = 'SEEK_ABSOLUTE'
SEEK_PERCENTAGE = 'SEEK_PERCENTAGE'

ASPECT_ORIGINAL = 'ASPECT_ORIGINAL'
ASPECT_16_9 = 'ASPECT_16_9'
ASPECT_4_3 = 'ASPECT_4_3'
ASPECT_ZOOM = 'ASPECT_ZOOM'

class Player(candy.Widget):
    """
    Gstreamer video widget.
    """

    state = gst.State.NULL
    aspect = original_aspect = None
    stream_changed = False
    zoom = 1

    def create(self):
        """
        Create the clutter object
        """
        self.delayed = []
        self.obj = clutter.Texture()
        self.obj.hide()

    def update(self, modified):
        """
        Render the widget
        """
        if 'width' in modified or 'height' in modified:
            if 'width' in modified and self.width:
                self.obj.set_width(self.width)
            if 'height' in modified and self.height:
                self.obj.set_height(self.height)
            self.calculate_geometry()
        if 'uri' in modified and self.uri:
            self.pipeline = gst.ElementFactory.make('playbin', 'pipeline')
            # There is a memory problem somehow. If we loose the sink
            # here gstreamer does not has a ref on its own and it
            # crashes. If we just keep the ref until the widget is
            # destroyed, it dies at that point (or later when we
            # repeat this). Even storing the sink forever crashes
            # gstreamer with gobject messages at some point. Maybe I'm
            # doing something wrong here, but I don't think so.
            sink = gst.ElementFactory.make('cluttersink', 'video')
            sink.set_property('texture', self.obj)
            self.pipeline.set_property('video-sink', sink)
            # monitor the message bus
            bus = self.pipeline.get_bus()
            bus.connect('message', self.event_message)
            bus.add_signal_watch()
            if self.uri.endswith('/'):
                self.uri = self.uri[:-1]
            if self.uri.find('://') == -1:
                self.uri = 'file://' + self.uri
            self.pipeline.set_property("uri", self.uri)
            self.streaminfo = {
                'audio': {},
                'subtitle': {},
                'is_menu': False,
                'sync': True,
                'current-audio': 0,
                'current-subtitle': -1
            }
            if self.uri.startswith('file://'):
                metadata = kaa.metadata.parse(self.uri)
                if metadata and 'audio' in metadata:
                    # video item, not audio only
                    for audio in metadata.audio:
                        self.streaminfo['audio'][audio.id] = None if audio.langcode == 'und' else audio.langcode
                    for sub in metadata.subtitles:
                        self.streaminfo['subtitle'][sub.id] = None if sub.langcode == 'und' else sub.langcode
            self.send_widget_event('streaminfo', self.streaminfo)


    def calculate_geometry(self, secs=0):
        """
        Calculate geometry values based on requested geometry and aspect
        """
        if not self.aspect:
            return
        width, height = self.width * self.zoom, self.height * self.zoom
        if int(height * self.aspect) > width:
            height = int(width / self.aspect)
        else:
            width = int(height * self.aspect)
        x = self.x + int(self.width - width) / 2
        y = self.y + int(self.height - height) / 2
        self.obj.animatev(clutter.AnimationMode.EASE_IN_QUAD, int(secs * 1000) or 1,
             ['x', 'y', 'width', 'height'], [x,y,width, height])

    #
    # control callbacks from the main process
    #

    def do_play(self):
        """
        Start playback
        """
        if not self.pipeline:
            return
        self.pipeline.set_state(gst.State.PLAYING)
        gobject.timeout_add(200, self.event_progress)

    def do_stop(self):
        """
        Stop playback
        """
        if not self.pipeline:
            return
        self.pipeline.set_state(gst.State.NULL)
        self.pipeline = None
        self.send_widget_event('finished')

    @requires_state(gst.State.PLAYING, gst.State.PAUSED)
    def do_pause(self):
        """
        Pause playback
        """
        if not self.pipeline:
            return
        self.pipeline.set_state(gst.State.PAUSED)

    @requires_state(gst.State.PLAYING, gst.State.PAUSED)
    def do_resume(self):
        """
        Resume playback
        """
        if not self.pipeline:
            return
        self.pipeline.set_state(gst.State.PLAYING)

    @requires_state(gst.State.PLAYING, gst.State.PAUSED)
    def do_seek(self, value, type):
        """
        Seek to the given position
        """
        if not self.pipeline:
            return
        duration = self.pipeline.query_duration(gst.Format.TIME)[1]
        if not duration:
            return
        if type == SEEK_RELATIVE:
            current = self.pipeline.query_position(gst.Format.TIME)[1]
            pos = current + (value * gst.SECOND)
        if type == SEEK_ABSOLUTE:
            pos = value * gst.SECOND
        if type == SEEK_PERCENTAGE:
            pos = (duration * value) / 100.0
        self.pipeline.seek_simple(gst.Format.TIME, gst.SeekFlags.FLUSH | gst.SeekFlags.KEY_UNIT, min(pos, duration))

    @requires_state(gst.State.PLAYING)
    def do_set_audio(self, idx):
        """
        Set the audio stream
        """
        if not self.pipeline:
            return
        self.pipeline.set_property('current-audio', idx)
        self.stream_changed = True

    @requires_state(gst.State.PLAYING)
    def do_set_subtitle(self, idx):
        """
        Set the subtitle stream (-1 == none)
        """
        if not self.pipeline:
            return
        flags = self.pipeline.get_property('flags')
        if flags == flags | GST_PLAY_FLAG_TEXT and idx == -1:
            self.pipeline.set_property('flags', flags ^ GST_PLAY_FLAG_TEXT)
        if flags != flags | GST_PLAY_FLAG_TEXT and idx >= 0:
            self.pipeline.set_property('flags', flags | GST_PLAY_FLAG_TEXT)
        if idx >= 0:
            self.pipeline.set_property('current-text', idx)
        self.stream_changed = True

    def do_set_aspect(self, aspect):
        """
        Set the aspect ratio
        """
        if aspect == ASPECT_ORIGINAL:
            self.zoom = 1
            self.aspect = self.original_aspect
        if aspect == ASPECT_16_9:
            self.zoom = 1
            self.aspect = 16.0 / 9
        if aspect == ASPECT_4_3:
            self.zoom = 1
            self.aspect = 4.0 / 3
        if aspect == ASPECT_ZOOM:
            self.aspect = 4.0 / 3
            self.zoom = 4.0 / 3
        self.calculate_geometry(0.2)

    def do_set_deinterlace(self, value):
        """
        Turn on/off deinterlacing
        """
        pass

    def do_nav_command(self, cmd):
        """
        Send DVD navigation command
        """
        pass

    #
    # events from gstreamer
    #

    def event_message(self, bus, message):
        """
        Events from GStreamer
        """
        if message.type == gst.MessageType.STATE_CHANGED:
            state = message.parse_state_changed()
            if self.state in (gst.State.PLAYING, gst.State.PAUSED):
                # toggle between play/pause
                state = state[1]
            elif state[1] in (gst.State.PLAYING, gst.State.PAUSED):
                # initial stream start
                state_change_return, state = self.pipeline.get_state(0)[:2]
                if state_change_return != gst.StateChangeReturn.SUCCESS:
                    # invalid state
                    return
            else:
                # invalid state
                return
            if self.state != state:
                previous, self.state = self.state, state
                self.event_state_change(previous, state)
        elif message.type == gst.MessageType.ERROR:
            log.error(str(message.parse_error()[1]))
            self.do_stop()
        elif message.type == gst.MessageType.EOS:
            self.send_widget_event('finished')

    def event_state_change(self, previous, state):
        """
        Gstreamer state change
        """
        for delayed in self.delayed[:]:
            if self.state in delayed[0]:
                func, args, kwargs = delayed[1:]
                func(self, *args, **kwargs)
                self.delayed.remove(delayed)
        if previous == gst.State.NULL:
            caps = self.pipeline.emit('get-video-pad', 0)
            s = caps.get_current_caps().get_structure(0)
            width = s.get_int('width')[1]
            height = s.get_int('height')[1]
            # Resize and move the texture to keep the aspect
            # ratio. This is kind of ugly because when the application
            # changes it this calculation won't happen again. And I
            # guess we should also connect to cap changes to redo this
            # calculations. Besides that, it should be possible to
            # change the aspect ratio during playback time,
            # e.g. ignore it and zoom.
            a1, a2 = s.get_fraction('pixel-aspect-ratio')[1:]
            aspect = (float(width) / height) * (float(a1) / a2)
            if int(self.height * aspect) > self.width:
                height = int(self.width / aspect)
                self.obj.set_y(self.y + int(self.height - height) / 2)
                self.obj.set_height(height)
            else:
                width = int(self.height * aspect)
                self.obj.set_x(self.x + int(self.width - width) / 2)
                self.obj.set_width(width)
            # Something else that could be interessting in the future:
            # s.get_boolean('interlaced')
            self.obj.show()
            # turn off subtitles by default
            self.do_set_subtitle(-1)

    def event_size_change(self, texture, base_width, base_height):
        """
        Video size change
        """
        if self.audio_only and self.visualisation:
            return
        self.aspect = self.original_aspect = float(base_width) / base_height
        self.zoom = 1
        self.calculate_geometry()

    def event_progress(self):
        """
        Progress update
        """
        if not self.pipeline:
            return False
        if not self.state in (gst.State.PLAYING, gst.State.PAUSED):
            return True
        pos = float(self.pipeline.query_position(gst.Format.TIME)[1]) / 1000000000
        self.send_widget_event('progress', pos)
        if self.stream_changed:
            self.stream_changed = False
            if self.pipeline.get_property('flags') == self.pipeline.get_property('flags') | GST_PLAY_FLAG_TEXT:
                self.streaminfo['current-subtitle'] = self.pipeline.get_property('current-text')
            else:
                self.streaminfo['current-subtitle'] = -1
            self.streaminfo['current-audio'] = self.pipeline.get_property('current-audio')
            self.send_widget_event('streaminfo', self.streaminfo)
        return True



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# candyxml.py - Parser to parse XML into widget Templates
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'parse', 'register', 'get_class' ]

# python imports
import os
import logging
import xml.sax
import imp

# kaa.candy imports
from core import Color, Font

# get logging object
log = logging.getLogger('kaa.candy')


class ElementDict(dict):

    def __getattr__(self, attr):
        return self.get(attr)

    def remove(self, *attrs):
        for attr in attrs:
            super(ElementDict, self).remove(attr)
        return self

class Element(object):
    """
    XML node element.
    """
    def __init__(self, node, parent, attrs):
        self.content = ''
        self.node = node
        # note: circular reference
        self.parent = parent
        self._attrs = {}
        for key, value in attrs.items():
            if key in ('x', 'xpadding', 'y', 'ypadding'):
                value = int(value)
            elif key == 'width' and not value.endswith('%'):
                x1 = int(attrs.get('x', 0))
                x2 = int(attrs.get('x', 0)) + int(value)
                value = x2 - x1
            elif key == 'height' and not value.endswith('%'):
                y1 = int(attrs.get('y', 0))
                y2 = int(attrs.get('y', 0)) + int(value)
                value = y2 - y1
            elif key in ('radius', 'size', 'spacing'):
                value = int(value)
            elif key.find('color') != -1:
                value = Color(value)
            elif key.find('font') != -1:
                value = Font(value)
                value.size = int(value.size)
            self._attrs[str(key).replace('-', '_')] = value
        self._children = []

    @property
    def candyxml(self):
        root = self.parent
        while getattr(root, 'parent', None) is not None:
            root = root.parent
        return root

    def __iter__(self):
        """
        Iterate over the list of children.
        """
        return self._children[:].__iter__()

    def __getitem__(self, pos):
        """
        Return nth child.
        """
        return self._children[pos]

    def __getattr__(self, attr):
        """
        Return attribute or child with the given name.
        """
        if attr == 'pos':
            return [ self._attrs.get('x', 0), self._attrs.get('y', 0) ]
        if attr == 'size':
            return self.width, self.height
        value = self._attrs.get(attr)
        if value is not None:
            return value
        for child in self._children:
            if child.node == attr:
                return child
        if attr in ('width', 'height'):
            # Set width or height to None. All widgets except the grid will
            # accept such values. The real value will be inserted later
            # based on the parent settings
            return None
        return None

    def xmlcreate(element):
        """
        Create a template or object from XML.
        """
        parser = _parser.get(element.node)
        if parser is None:
            raise RuntimeError('no parser for %s' % element.node)
        parser = parser.candyxml_parse(element)
        if parser is None:
            raise RuntimeError('no parser for %s:%s' % (element.node, element.style))
        return getattr(parser, '__template__', parser).candyxml_create(element)

    def get_children(self, node=None):
        """
        Return all children with the given node name
        """
        if node is None:
            return self._children[:]
        return [ c for c in self._children if c.node == node ]

    def attributes(self):
        """
        Get key/value list of all attributes.,
        """
        return self._attrs.items()

    def remove(self, child):
        """
        Remove the given child element.
        """
        self._children.remove(child)


class CandyXML(xml.sax.ContentHandler):
    """
    candyxml parser.
    """
    def __init__(self, data, elements=None, path=None):
        xml.sax.ContentHandler.__init__(self)
        self._elements = elements or ElementDict()
        # Internal stuff
        self._root = None
        self._current = None
        self._stack = []
        self._names = []
        self._path = path
        self.scripts = {}
        self._parser = xml.sax.make_parser()
        self._parser.setContentHandler(self)
        if data.find('<') >= 0:
            # data is xml data
            self._parser.feed(data)
        else:
            # data is filename
            self._parser.parse(data)

    def get_elements(self):
        """
        Return root attributes and parsed elements
        """
        if self._root[0] == '__candyxml_simple__':
            # fake candyxml tree, only one element
            return self._elements.values()[0].values()[0]
        return dict(self._root[1]), self._elements

    def startElement(self, name, attrs):
        """
        SAX Callback.
        """
        if self._root is None:
            self._root = name, attrs
            if not name in _parser.keys():
                # must be a parent tag like cnadyxml around
                # everything. This means we may have more than one
                # widget in this xml stream.
                return
            # create fake root
            self._root = '__candyxml_simple__', {}
        if name == 'alias' and len(self._stack) == 0:
            self._names.append(attrs['name'])
            return
        element = Element(name, self._current or self, attrs)
        if self._current is not None:
            self._stack.append(self._current)
            self._current._children.append(element)
        else:
            self._names.append(attrs.get('name'))
        self._current = element

    def endElement(self, name):
        """
        SAX Callback.
        """
        if self._current:
            if self._current.content.strip().find('\n') == -1:
                self._current.content = self._current.content.strip()
        if len(self._stack):
            self._current = self._stack.pop()
        elif name == 'alias':
            # alias for high level element, skip
            return
        elif name == 'script':
            if self._path:
                name = os.path.splitext(self._current.filename)[0]
                (file, filename, data) = imp.find_module(name, [ self._path ])
                module = imp.load_module(name, file, filename, data)
                for m in dir(module):
                    if not m.startswith('_'):
                        self.scripts[m] = getattr(module, m)
            self._current = None
            self._names = []
        elif name != self._root[0]:
            screen = self._current.xmlcreate()
            if not self._elements.get(name):
                self._elements[name] = ElementDict()
            for subname in self._names:
                self._elements[name][subname] = screen
            self._current = None
            self._names = []

    def characters(self, c):
        """
        SAX callback
        """
        if self._current:
            self._current.content += c


def parse(data, elements=None):
    """
    Load a candyxml file based on the given screen resolution.
    @param data: filename of the XML file to parse or XML data
    @returns: root element attributes and dict of parsed elements
    """
    if not os.path.isdir(data):
        return CandyXML(data, elements).get_elements()
    attributes = {}
    for f in os.listdir(data):
        if not f.endswith('.xml'):
            continue
        f = os.path.join(data, f)
        try:
            a, elements = CandyXML(f, elements, data).get_elements()
            attributes.update(a)
        except Exception, e:
            log.exception('parse error in %s', f)
    return attributes, elements

class Styles(dict):
    """
    Style dict for candyxml_parse and candyxml_create callbacks
    """
    class __metaclass__(type):
        def __new__(meta, name, bases, attrs):
            cls = type.__new__(meta, name, bases, attrs)
            if 'candyxml_name' in attrs.keys():
                register(cls())
            return cls

    def candyxml_parse(self, element):
        return self.get(element.style)

#: list of candyxml parser
_parser = {}

def register(cls):
    """
    Register a class
    """
    name = cls.candyxml_name
    parser = _parser
    if not isinstance(cls, Styles):
        if not name in parser:
            parser[name] = Styles()
        parser, name = parser[name], getattr(cls, 'candyxml_style', None)
    if name in parser and not cls.candyxml_override:
        log.info('%s already registered' % name)
    parser[name] = cls

def get_class(name, style):
    """
    Get the class registered to the given name and style
    """
    return _parser.get(name).get(style)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# text.py - multiline text widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Text' ]

import re
import kaa.base

import cairo

from gi.repository import Pango, PangoCairo

from widget import Widget
from ..core import Color, Font, create_cairo_context


class Text(Widget):
    candyxml_name = 'text'
    candy_backend = 'candy.Text'
    attributes = [ 'color', 'font', 'text', 'align' ]
    attribute_types = {
        'color': Color,
        'font': Font
    }

    __intrinsic_size_param = None
    __intrinsic_size_cache = None

    __text_regexp = re.compile('\$([a-zA-Z][a-zA-Z0-9_\.]*)|\${([^}]*)}')
    __text = None

    def __init__(self, pos, size, text, font, color, align=None, context=None):
        """
        Create Text widget. Unlike a Label a Text widget supports multi-line
        text and markup. See the pango markup documentation.

        @param pos: (x,y) position of the widget or None
        @param size: (width,height) geometry of the widget
        @param text: text to show
        @param color: kaa.candy.Color to fill the text
        @param align: align value
        @param context: the context the widget is created in
        """
        super(Text, self).__init__(pos, size, context)
        self.align = align or Widget.ALIGN_LEFT
        self.font = font
        self.text = text
        self.color = color

    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, text):
        self.__text_provided = text
        def replace_context(matchobj):
            match = matchobj.groups()[0] or matchobj.groups()[1]
            s = self.context.get(match, '')
            if s is not None:
                return kaa.base.py3_str(s, coerce=True)
            return ''
        if self.context:
            # we have a context, use it
            text = re.sub(self.__text_regexp, replace_context, text)
        if self.__text == text:
            return
        self.__text = text
        self.queue_rendering()

    def sync_context(self):
        """
        Adjust to a new context
        """
        self.text = self.__text_provided

    def sync_layout(self, size):
        """
        Sync layout changes and calculate intrinsic size based on the
        parent's size.
        """
        super(Text, self).sync_layout(size)
        width, height = self.size
        if self.__intrinsic_size_param == (width, height, self.text, self.font.name, self.font.size):
            self.intrinsic_size = self.__intrinsic_size_cache
            return self.__intrinsic_size_cache
        cr = create_cairo_context()
        layout = PangoCairo.create_layout(cr)
        layout.set_width(width * Pango.SCALE)
        layout.set_height(height * Pango.SCALE)
        layout.set_ellipsize(Pango.EllipsizeMode.END)
        layout.set_wrap(Pango.WrapMode.WORD_CHAR)
        layout.set_font_description(self.font.get_font_description())
        layout.set_text(self.text, -1)
        PangoCairo.show_layout(cr, layout)
        self.__intrinsic_size_cache = \
            int(min(width, Pango.units_to_double(layout.get_size()[0]))), \
            int(min(height, Pango.units_to_double(layout.get_size()[1])))
        self.__intrinsic_size_param = (width, height, self.text, self.font.name, self.font.size)
        self.intrinsic_size = self.__intrinsic_size_cache
        return self.__intrinsic_size_cache

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the XML element for parameter to create the widget.
        """
        return super(Text, cls).candyxml_parse(element).update(
            text=element.content, align=element.align, color=element.color,
            font=element.font)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# video.py - video widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2012 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Video', 'Audio', 'SEEK_RELATIVE', 'SEEK_ABSOLUTE', 'SEEK_PERCENTAGE',
            'STATE_IDLE', 'STATE_PLAYING', 'STATE_PAUSED', 'NEXT', 'POSSIBLE_PLAYER' ]

# python imports
import logging

# kaa imports
import kaa
import kaa.metadata

# kaa.candy imports
from widget import Widget
from .. import config

# get logging object
log = logging.getLogger('kaa.candy')

SEEK_RELATIVE = 'SEEK_RELATIVE'
SEEK_ABSOLUTE = 'SEEK_ABSOLUTE'
SEEK_PERCENTAGE = 'SEEK_PERCENTAGE'

STATE_IDLE = 'STATE_IDLE'
STATE_PLAYING = 'STATE_PLAYING'
STATE_PAUSED = 'STATE_PAUSED'

ASPECT_ORIGINAL = 'ASPECT_ORIGINAL'
ASPECT_16_9 = 'ASPECT_16_9'
ASPECT_4_3 = 'ASPECT_4_3'
ASPECT_ZOOM = 'ASPECT_ZOOM'

ASPECTS = [ ASPECT_ORIGINAL, ASPECT_16_9, ASPECT_4_3, ASPECT_ZOOM ]

NEXT = 'NEXT'

# filled with values from the backend later
POSSIBLE_PLAYER = []

class Video(Widget):
    """
    Video widget
    """
    candyxml_name = 'video'
    candy_backend = 'candy.Video'

    attributes = [ 'uri', 'config', 'audio_only', 'player' ]
    audio_only = False

    __player = None

    def __init__(self, pos=None, size=None, uri=None, player='gstreamer', context=None):
        """
        Create the video widget. The widget supports gstreamer
        (default) and mplayer but only gstreamer can be used as real
        widget for now. When choosing mplayer it will always open a
        full screen window to play the video.

        The playback can be configured using the config member
        dictionary. Please note, that gstreamer tries to figure out
        most of the stuff itself and AC3 and DTS passthrough only
        works when using pulseaudio and pulseaudio configured
        correctly (pavucontrol). Future versions of kaa.candy may have
        more or changed options.
        """
        super(Video, self).__init__(pos, size, context)
        self.uri = uri
        self.signals = kaa.Signals('finished', 'progress', 'streaminfo')
        self.state = STATE_IDLE
        # player configuration
        self.config = {
            'mplayer.vdpau': False,
            'mplayer.passthrough': False,
            'fresh-rate': None
        }
        # current streaminfo / audio / subtitle values
        self.streaminfo = {
            'audio': {},
            'subtitle': {},
            'is_menu': False,
        }
        self.aspect = ASPECT_ORIGINAL
        self.player = player or 'gstreamer'

    @property
    def player(self):
        return self.__player

    @player.setter
    def player(self, value):
        if self.state != STATE_IDLE:
            raise RuntimeError('player already running')
        self.__player = value

    @property
    def uri(self):
        return self.__uri

    @uri.setter
    def uri(self, value):
        if value and isinstance(value, (str, unicode)) and value.startswith('$'):
            # variable from the context, e.g. $varname
            value = self.context.get(value) or ''
        if value and not value.find('://') > 0:
            value = 'file://' + value
        if value:
            self.metadata = kaa.metadata.parse(value)
        else:
            self.metadata = None
        self.__uri = value

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget.
        """
        return super(Video, cls).candyxml_parse(element).update(
            uri=element.uri or element.filename, player=element.player)

    #
    # public API to control the player
    #

    def play(self):
        """
        Start the playback
        """
        if self.state != STATE_IDLE:
            raise RuntimeError('player already running')
        if self.player not in POSSIBLE_PLAYER:
            raise RuntimeError('unknown player %s' % self.player)
        self.state = STATE_PLAYING
        self.backend.do_play()

    def stop(self):
        """
        Stop the playback
        """
        if self.state != STATE_IDLE:
            self.backend.do_stop()

    def pause(self):
        """
        Pause playback
        """
        if self.state == STATE_PLAYING:
            self.backend.do_pause()
            self.state = STATE_PAUSED

    def resume(self):
        """
        Resume a paused playback
        """
        if self.state == STATE_PAUSED:
            self.backend.do_resume()
            self.state = STATE_PLAYING

    def seek(self, value, type=SEEK_RELATIVE):
        """
        Seek to the given position. Type is either SEEK_RELATIVE
        (default), SEEK_ABSOLUTE or SEEK_PERCENTAGE.
        """
        self.backend.do_seek(value, type)

    def set_audio(self, idx):
        """
        Set the audio channel to stream number idx
        """
        self.backend.do_set_audio(idx)
        return idx

    def set_subtitle(self, idx):
        """
        Set the subtitle sream idx. Use -1 to turn subtitles off.
        """
        self.backend.do_set_subtitle(idx)
        return idx

    def set_deinterlace(self, value):
        """
        Turn on/off deinterlacing
        """
        self.backend.do_set_deinterlace(value)

    def set_aspect(self, aspect):
        """
        Set the aspect ratio
        """
        if aspect == NEXT:
            aspect = ASPECTS[(ASPECTS.index(self.aspect) + 1) % len(ASPECTS)]
        self.backend.do_set_aspect(aspect)
        self.aspect = aspect

    def nav_command(self, cmd):
        """
        Send DVD navigation command
        """
        self.backend.do_nav_command(cmd)

    #
    # backend callbacks
    #

    def event_progress(self, pos):
        """
        Callback from the backend: new progress information
        """
        self.signals['progress'].emit(pos)

    def event_finished(self):
        """
        Callback from the backend: playback finished
        """
        self.state = STATE_IDLE
        self.signals['finished'].emit()

    def event_streaminfo(self, streaminfo):
        """
        Callback from the backend: streaminfo
        """
        del streaminfo['sync']
        self.signals['streaminfo'].emit(streaminfo)
        self.streaminfo = streaminfo

class Audio(Video):
    """
    Hidden video widget for audio only
    """
    candyxml_name = 'audio'
    candy_backend = 'candy.Audio'
    audio_only = True

    attributes = Video.attributes + [ 'visualisation' ]

    def __init__(self, pos=None, size=None, uri=None, player='gstreamer', visualisation=None,
                 context=None):
        """
        Create the audio widget. If visualisation is None it is invisible.
        """
        super(Audio, self).__init__(pos, size, uri, player, context)
        self.visualisation = visualisation

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget.
        """
        return super(Audio, cls).candyxml_parse(element).update(
            visualisation=element.visualisation)



import os
import gc
import sys
import kaa
import kaa.candy

xml = '''
<candyxml geometry="800x600">
    <group name="g1" x="10" y="10" width="600" height="580">
        <rectangle width="50%" height="100" color="0xffff00" radius="50"
           border_size="5" border_color="0xffffff"/>
        <label y="120" font="Vera:24" color="0x00ff00">Test</label>
        <text y="150" font="Vera:24" color="0xff0000" align="center">
Multiline
Text
        </text>
        <label y="180" font="Vera:24" color="0x00ff00" width="200" height="100">
            <properties xalign="center" yalign="center"/>
            center
        </label>
        <group name="g2" x="10" y="300">
            <rectangle x="10" width="100%" y="10" height="100%" color="0x08cccccc" radius="50"
               border_size="5" border_color="0xffffff">
                <properties reference_x="siblings" reference_y="siblings"/>
            </rectangle>
            <rectangle x="20" width="100" height="100" color="0xffff00" radius="50"
               border_size="5" border_color="0xffffff"/>
        </group>
        <group name="g3">
            <properties xalign="center" yalign="bottom"/>
            <rectangle x="10" width="100%" y="10" height="100%" color="0x05cccccc">
                <properties reference_x="siblings" reference_y="siblings"/>
            </rectangle>
            <label x="20" y="20" color="0x000000" font="Vera:24">highlight
                <properties xalign="shrink" yalign="shrink"/>
            </label>
        </group>
        <image y="400" url="test/logo.png" width="200" height="200">
            <properties keep_aspect="True" yalign="center"/>
        </image>
    </group>
</candyxml>
'''

@kaa.coroutine()
def main():
    stage = kaa.candy.Stage((400,300), 'candy')
    attr, candy = stage.candyxml(xml)
    container = candy.group.g1()
    stage.add(container)
    yield kaa.NotFinished
    yield kaa.NotFinished
    container.x += 100
    yield kaa.NotFinished
    #container.parent = None
    yield kaa.NotFinished

def garbage_collect():
    gc.collect()
    for g in gc.garbage:
        print 'Unable to free %s' % g

# Set up garbage collector to be called every 5 seconds
kaa.Timer(garbage_collect).start(1)

main()
kaa.main.run()
print 'shutdown'



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# core.py - Core classes
# -----------------------------------------------------------------------------
# Note: this file is imported from the application using kaa.candy as
# well as the rendering process. Therefore, no imports from this file
# to other parts of kaa.candy are allowed to avoid strange side
# effects.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Context', 'Color', 'Font' ]

# python imports
import logging
import cairo

from gi.repository import Pango

# kaa imports
import kaa

# get logging object
log = logging.getLogger('kaa.candy')

class Context(dict):

    def __init__(self, ctx=None, **kwargs):
        if ctx is not None:
            super(Context, self).__init__(ctx.copy())
        self.update(kwargs)

    def get(self, attr, default=None):
        if attr.startswith('$'):
            # strip prefix for variables if set
            attr = attr[1:]
        try:
            # try the variable as it is
            value = eval(attr, self)
            return value
        except Exception, e:
            log.error('unable to evaluate %s', attr)
            return default

    def __getattr__(self, attr):
        return self.get(attr)


class Color(list):
    """
    Color object which is a list of r,g,b,a with values between 0 and 255.
    """
    def __init__(self, *col):
        """
        Create a new color object. All C{set_color} member functions of kaa.candy
        widgets use this class for setting a color and not the clutter color object.
        The Color object is list of r,g,b,a with values between 0 and 255.

        @param col: one of the following types
         - a tuple r,g,b,a
         - a clutter color object
         - a string #aarrggbb
        """
        if len(col) > 1:
            return super(Color, self).__init__(col)
        col = col[0]
        if col == None:
            return super(Color, self).__init__((0,0,0,255))
        if hasattr(col, 'red'):
            # clutter.Color object
            return super(Color, self).__init__((col.red, col.green, col.blue, col.alpha))
        if isinstance(col, (list, tuple)):
            # tuple as one argument
            return super(Color, self).__init__(col)
        # Convert a 32-bit ARGB string 0xaarrggbb
        if not isinstance(col, (int, long)):
            col = long(col, 16)
        a = 255 - ((col >> 24) & 0xff)
        r = (col >> 16) & 0xff
        g = (col >> 8) & 0xff
        b = (col >> 0) & 0xff
        super(Color, self).__init__((r,g,b,a))

    def to_cairo(self):
        """
        Convert to list used by cairo.

        @returns: list with float values from 0 to 1.0
        """
        return [ x / 255.0 for x in self ]


# helping cairo surface
_font_cairo_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)

def create_cairo_context():
    """
    Create a dummy cairo context object for font calculations.
    """
    return cairo.Context(_font_cairo_surface)


class Font(object):
    """
    Font object containing font name and font size

    @ivar name: font name
    @ivar size: font size
    """

    __height_cache = {}

    ASCENT, TYPICAL, MAX_HEIGHT = range(3)

    def __init__(self, name):
        """
        Create a new font object
        @param name: name and size of the font, e.g. Vera:24
        """
        self.name = name
        self.size = 0
        if name.find(':') > 0:
            self.name, size = name.split(':')
            self.size = int(size)

    def get_height(self, field=None, size=None):
        """
        Get height of a text with this font.
        @returns: ascent (typical height above the baseline), normal (typical height
            of a string without special characters) and ascent + descent
        """
        if size is None:
            size = self.size
        info = self.__height_cache.get((self.name, size))
        if info is None:
            c = create_cairo_context()
            c.select_font_face(self.name, cairo.FONT_SLANT_NORMAL)
            c.set_font_size(size)
            ascent, descent = c.font_extents()[:2]
            info = int(ascent), int(-c.text_extents(u'Ag')[1]), int(ascent + descent)
            self.__height_cache[(self.name, size)] = info
        if field is None:
            return info
        return info[field]

    def get_width(self, text):
        """
        Get width of the given string
        """
        c = create_cairo_context()
        c.select_font_face(self.name, cairo.FONT_SLANT_NORMAL)
        c.set_font_size(self.size)
        # add x_bearing to width (maybe use x_advance later)
        # http://cairographics.org/manual/cairo-Scaled-Fonts.html#cairo-text-extents-t
        ext = c.text_extents(text)
        return int(ext[0] + ext[2]) + 1

    def get_font(self, height):
        """
        Get font object with size set to fit the given height.
        """
        size = 1
        last = None
        while True:
            if not self.__height_cache.get((self.name, size)):
                self.get_height(size=size)
            if self.__height_cache.get((self.name, size))[Font.MAX_HEIGHT] > height:
                f = Font(self.name)
                f.size = size - 1
                return f
            last = size
            size += 1

    def get_font_description(self):
        """
        Get the Pango.FontDescription for the font object
        """
        return Pango.FontDescription.from_string("%s %spx" % (self.name, self.size))



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# widget.py - base widget class
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2013 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'XMLdict', 'Widget' ]

# Python imports
import os
import sys
import logging

# kaa imports
import kaa
import kaa.weakref

# kaa.candy imports
from .. import candyxml
from ..core import Context
from ..template import Template

# get logging object
log = logging.getLogger('kaa.candy')


next_id = 1

NOT_SET = object()

class XMLdict(dict):
    """
    XML parser dict helper class.
    """
    candyname = None
    def update(self, **kwargs):
        super(XMLdict, self).update(**kwargs)
        return self
    def remove(self, *args):
        for arg in args:
            self.pop(arg)
        return self

class BackendWrapper(object):
    def __init__(self, candy_id):
        self.candy_id = candy_id
        self.queue = []
        self.stage = None

    def call(self, cmd, *args):
        if self.stage:
            return self.stage.queue_command(self.candy_id, cmd, args)
        self.queue.append((self.candy_id, cmd, args))

    def __getattr__(self, attr):
        return lambda *args: self.call(attr, *args)


class Widget(object):

    candy_backend = 'candy.Widget'
    candyxml_name = None
    candyxml_style = None
    candyxml_override = False

    attributes = []
    attribute_types = {}

    ALIGN_LEFT = 'left'
    ALIGN_RIGHT = 'right'
    ALIGN_TOP = 'top'
    ALIGN_BOTTOM = 'bottom'
    ALIGN_CENTER = 'center'
    ALIGN_SHRINK = 'shrink'

    # internal class variables
    _candy_sync_new = []
    _candy_sync_delete = []
    _candy_sync_reparent = []
    _candy_import = []
    _candy_backends = {}
    _candy_all_widgets = []

    # internal object variables
    _candy_id = None
    _candy_dirty = True
    _candy_stack = None
    stage = None

    class __metaclass__(type):
        def __new__(meta, name, bases, attrs):
            cls = type.__new__(meta, name, bases, attrs)
            if 'candyxml_name' in attrs.keys() or 'candyxml_style' in attrs.keys() \
                    or 'candyxml_override' in attrs.keys():
                candyxml.register(cls)
            backend = cls.candy_backend.split('.')[0]
            if backend != 'candy':
                path = os.path.dirname(sys.modules[cls.__module__].__file__) + '/' + backend
                if not path in cls._candy_backends:
                    dyn = 'dyn%03d' % len(cls._candy_backends.keys())
                    cls._candy_backends[path] = dyn
                    cls._candy_import.append((dyn, path))
                cls.candy_backend = cls._candy_backends[path] + cls.candy_backend[len(backend):]
            return cls

    #: template for object creation
    __template__ = Template

    # widgets with dynamic size depend on the size of the paent or
    # other widgets in a group based on the reference setting.
    # Possible values are 'parent' based on the parents geometry and
    # 'siblings' based on its siblings.
    reference_x = 'parent'
    reference_y = 'parent'

    # the geometry values depend on some internal calculations.
    # Therefore, they are hidden using properties.
    __x = 0
    __y = 0
    __width = None
    __height = None
    __intrinsic_size = None
    __variable_width = 100
    __variable_height = 100

    __parent = None
    __stage = None

    # attributes
    name = None

    xalign = None
    yalign = None

    opacity = 255
    scale_x = 1
    scale_y = 1
    anchor_point = 0, 0

    visible = True

    # if True, the context will not be updated by the parent
    freeze_context = False

    def __init__(self, pos=None, size=None, context=None):
        self.attributes = self.attributes + ['xalign', 'yalign', 'opacity', 'scale_x', 'scale_y', 'anchor_point', 'visible']
        self._candy_events = {}
        global next_id
        self._candy_id = next_id
        next_id += 1
        self.backend = BackendWrapper(self._candy_id)
        self.__weakref = kaa.weakref.weakref(self)
        Widget._candy_all_widgets.append(self.__weakref)
        Widget._candy_sync_new.append(self)
        if pos is not None:
            self.x, self.y = pos
        if size is not None:
            self.width, self.height = size
        self.__sync_cache = {}
        self.__context = context or {}
        self.__depends = {}

    def __setattr__(self, attr, value):
        if value and attr in self.attribute_types and not isinstance(value, self.attribute_types[attr]):
            value = self.attribute_types[attr](value)
        super(Widget, self).__setattr__(attr, value)
        if not self._candy_dirty and (attr in self.attributes):
            self.queue_rendering()

    def __del__(self):
        if not hasattr(self, '_candy_id') or not Widget:
            # Python is shutting down, no need to clean anymore -- and
            # not possible.
            return
        Widget._candy_all_widgets.remove(self.__weakref)
        Widget._candy_sync_delete.append(self._candy_id)
        if self.__stage and not self.__stage._candy_dirty:
            self.__stage.queue_rendering()

    def emit(self, event, *args):
        """
        Emit the given event
        """
        if event in self._candy_events:
            return self._candy_events[event](*args) or True
        return False

    def __reset__(self):
        """
        Internal function when the candy backend becomes invalid and
        needs to be restarted.
        """
        if not self in Widget._candy_sync_new:
            Widget._candy_sync_new.append(self)
        if not self in Widget._candy_sync_reparent and self.parent:
            Widget._candy_sync_reparent.append(self)
        self._candy_dirty = True
        self.__stage = None
        self.__sync_cache = {}

    def __sync__(self, tasks):
        """
        Internal function to add the changes to the list of tasks for
        the backend.
        """
        if not self._candy_dirty:
            return False
        if self._candy_events.get('create'):
            self._candy_events.pop('create')(self)
        (x, y), (width, height) = self.intrinsic_geometry
        # check the position and set a new position on the backend if
        # needed. This does not result in a new rendering.
        attributes = {}
        for (attr, value) in [ ('x', x), ('y', y) ]:
            if self.__sync_cache.get(attr, NOT_SET) != value:
                attributes[attr] = value
        if attributes:
            self.__sync_cache.update(attributes)
            tasks.append(('position', (self._candy_id, x, y)))
        # check all other attributes and this will cause a
        # re-rendering. Even width and height change the widget on the
        # backend.
        attributes = {}
        for attr in self.attributes:
            new_value = getattr(self, attr)
            if self.__sync_cache.get(attr, NOT_SET) != new_value:
                attributes[attr] = new_value
        for (attr, value) in [ ('width', width), ('height', height) ]:
            if self.__sync_cache.get(attr, NOT_SET) != value:
                attributes[attr] = value
        if attributes:
            self.__sync_cache.update(attributes)
            tasks.append(('update', (self._candy_id, attributes)))
        self._candy_dirty = False
        return True

    def queue_rendering(self):
        """
        Queue sync
        """
        self.__intrinsic_size = None
        # if self._candy_dirty:
        #     return True
        self._candy_dirty = True
        parent = self.parent
        if parent:# and not parent._candy_dirty:
            parent.queue_rendering()
        return False

    def sync_context(self):
        """
        Adjust to a new context
        """
        pass

    def sync_layout(self, (width, height)):
        """
        Sync layout changes and calculate intrinsic size based on the
        parent's size.
        """
        if self.__variable_width:
            self.__width = int((width * self.__variable_width) / 100)
        if self.__variable_height:
            self.__height = int((height * self.__variable_height) / 100)
        self.__intrinsic_size = self.__width, self.__height

    def sync_prepare(self):
        """
        Prepare widget for the next sync with the backend
        """
        return self._candy_dirty

    @kaa.coroutine()
    def animate(self, ease, secs, unparent=False, **kwargs):
        args = [item for sublist in kwargs.items() for item in sublist]
        if secs:
            self.backend.animate(ease, secs, *args)
            yield kaa.delay(secs)
            for key, value in kwargs.items():
                self.__sync_cache[key] = value
        for key, value in kwargs.items():
            setattr(self, key, value)
        if unparent:
            self.parent = None

    def add_dependencies(self, *vars):
        """
        Evaluate the context for the given variable and depend on the result

        :param var: variable name to eval
        """
        for var in vars:
            self.__depends[var] = repr(self.__context.get(var))

    def supports_context(self, context):
        """
        Check if the widget is capable of the given context based on its
        dependencies.

        :param context: context dict
        :returns: False if the widget can not handle the context or True
        """
        try:
            for var, value in self.__depends.items():
                if value != repr(context.get(var)):
                    return False
        except AttributeError, e:
            log.error('supports_context: %s', e)
            return False
        return True

    def raise_top(self):
        """
        Raise widget to the top of the stack
        """
        if self.parent and self in self.parent.children:
            self.parent.children.remove(self)
            self.parent.children.append(self)
        self.backend.raise_top()

    def lower_bottom(self):
        """
        Lower widget to the bottom of the stack
        """
        if self.parent and self in self.parent.children:
            self.parent.children.remove(self)
            self.parent.children.insert(0, self)
        self.backend.lower_bottom()

    def above_sibling(self, sibling):
        """
        Move the widget above the given sibling
        """
        if self.parent and self.sibling.parent and self in self.parent.children:
            log.error('TODO: above_sibling not implemented for restart')
        self.backend.above_sibling('candy:widget:%s' % sibling._candy_id)

    def unparent(self):
        """
        Unparent the widget. Calls widget.parent = None
        """
        self.parent = None

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, x):
        self.__x = x
        if not self._candy_dirty:
            self.queue_rendering()

    @property
    def y(self):
        return self.__y

    @y.setter
    def y(self, y):
        self.__y = y
        if not self._candy_dirty:
            self.queue_rendering()

    @property
    def width(self):
        if self.__variable_width and not self.__intrinsic_size and self.parent:
            # start intrinsic size calculations
            self.intrinsic_size
        return self.__width

    @width.setter
    def width(self, width):
        self.__intrinsic_size = None
        if isinstance(width, (str, unicode)):
            # use percent values provided by the string
            self.__variable_width = int(width[:-1])
            self.__width = -1
        elif width is None:
            self.__variable_width = 100
            self.__width = -1
        else:
            self.__variable_width = None
            self.__width = width
        if not self._candy_dirty:
            self.queue_rendering()

    @property
    def height(self):
        if self.__variable_height and not self.__intrinsic_size and self.parent:
            # start intrinsic size calculations
            self.intrinsic_size
        return self.__height

    @height.setter
    def height(self, height):
        self.__intrinsic_size = None
        if isinstance(height, (str, unicode)):
            # use percent values provided by the string
            self.__variable_height = int(height[:-1])
            self.__height = -1
        elif height is None:
            self.__variable_height = 100
            self.__height = -1
        else:
            self.__variable_height = None
            self.__height = height
        if not self._candy_dirty:
            self.queue_rendering()

    @property
    def size(self):
        return self.width, self.height

    @property
    def variable_size(self):
        return self.__variable_width or self.__variable_height

    @property
    def intrinsic_size(self):
        if not self.__intrinsic_size:
            if (self.__variable_width or self.__variable_height) and not self.parent.__intrinsic_size:
                self.parent.intrinsic_size
            else:
                self.sync_layout(self.parent.size)
        return self.__intrinsic_size

    @intrinsic_size.setter
    def intrinsic_size(self, size):
        self.__intrinsic_size = size

    @property
    def intrinsic_geometry(self):
        """
        The actual geometry of the object
        """
        x, y = self.__x, self.__y
        width, height = self.intrinsic_size
        if self.xalign == Widget.ALIGN_CENTER:
            x += int((self.__width - width) / 2)
        if self.xalign == Widget.ALIGN_RIGHT:
            x += int(self.__width - width)
        if self.yalign == Widget.ALIGN_CENTER:
            y += int((self.__height - height) / 2)
        if self.yalign == Widget.ALIGN_BOTTOM:
            y += int(self.__height - height)
        return (x + self.anchor_point[0], y + self.anchor_point[1]), (width, height)

    @property
    def parent(self):
        return self.__parent

    @parent.setter
    def parent(self, parent):
        self._candy_stack = None
        if not self in Widget._candy_sync_reparent:
            self.queue_rendering()
        if self.__parent:
            self.__parent.children.remove(self)
        if parent:
            self.__parent = parent.__weakref
            self.__parent.children.append(self)
        else:
            self.__parent = None
        if not self in Widget._candy_sync_reparent:
            if parent:
                parent.queue_rendering()
            Widget._candy_sync_reparent.append(self)
            self.queue_rendering()

    @property
    def stage(self):
        if not self.__stage and self.__parent:
            return self.parent.stage
        return self.__stage

    @stage.setter
    def stage(self, stage):
        if self.__stage:
            raise AttributeError('stage is a read-only variable')
        self.__stage = stage

    @property
    def context(self):
        return self.__context

    @context.setter
    def context(self, context):
        if not isinstance(context, Context):
            context = Context(context)
        self.__context = context
        self.sync_context()

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the candyxml element for parameter to create the widget. This function
        must be overwitten by a subclass for the correct parsing. This class
        only parses pos and size and children with a use-us attribute::
          <widget x='10' y='20' width='100' height='50'>
              <widget x='0' y='0' use-as='content'/>
          </widget>

        This will return a dictionary with pos, size, and content.
        """
        parameter = XMLdict(pos=element.pos, size=(element.width, element.height))
        for child in element[:]:
            if child.use_as:
                widget = child.xmlcreate()
                if not widget:
                    log.error('unable to parse %s', child.node)
                else:
                    parameter[str(child.use_as)] = widget
                element.remove(child)
        parameter.candyname = element.name
        return parameter



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# main - Backend main process
# -----------------------------------------------------------------------------
# The candy server will run an RPC server in the kaa mainloop and will
# execute commands. A widget is created and prepared in the kaa main
# loop; everything else, including imports is done in the clutter
# thread.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

# python imports
import sys
import os
import time
import threading
import imp
import logging
import logging.config
import logging.handlers

# get logging object
log = logging.getLogger('candy')

# import GObject from gi.repository before jumping into kaa to force
# kaa.gobject to use the gi repository
from gi.repository import GObject as gobject

# kaa imports
import kaa
import kaa.rpc

# hack to provide the kaa.candy core to the widgets
import kaa.candy.core
sys.modules['core'] = kaa.candy.core

class Mainloop(object):
    """
    Clutter mainloop.
    """

    def run(self):
        # Import clutter only in the gobject thread
        # This function will be the running mainloop
        try:
            from gi.repository import Clutter as clutter
            clutter.threads_init()
            clutter.init([])
            clutter.main()
        except Exception, e:
            log.exception('unable to import clutter')
            return

    def quit(self):
        log.info('shutdown clutter mainloop')
        # Import clutter only in the gobject thread
        from gi.repository import Clutter as clutter
        clutter.main_quit()

    def imp(self, name, path, server):
        """
        Import inside the clutter thread
        """
        path = os.path.abspath(path)
        basename, dirname = os.path.basename(path), os.path.dirname(path)
        (file, filename, data) = imp.find_module(basename, [ dirname ])
        # hack to provide the module to other modules for import
        sys.modules[name] = imp.load_module(name, file, filename, data)
        globals()[name] = sys.modules[name]
        if hasattr(sys.modules[name], 'init'):
            sys.modules[name].init(server)

    def sync(self, queue, event):
        """
        Sync the changes with the clutter objects
        Executed inside the clutter thread
        """
        t0 = time.time()
        freeze = False
        while queue:
            func, args = queue.pop(0)
            if func == 'freeze':
                # Freeze the scene. From now on it is not allowed to
                # go back to the clutter main loop to keep animations
                # alive or the user will see a half-finished scene.
                freeze = True
                continue
            try:
                func(*args)
            except Exception, e:
                log.exception('sync error: %s%s', func, args)
            if not freeze and queue:
                sync_time = time.time() - t0
                if sync_time > 0.01:
                    # We should use the logging module somehow and get the
                    # logging info to the main process. Only print out the
                    # sync time if we cannot make 100fps.
                    log.warning('sync took %0.4f sec' % sync_time)
                    # For further debug to detect what function /
                    # widget is slow taht it took way too long:
                    # print 'last call', func, args
                if sync_time > 0.001:
                    # Not finished yet, but the scene is not frozen
                    # and we can go back to the clutter main loop to
                    # keep animations alive.
                    return True
        sync_time = time.time() - t0
        if sync_time > 0.01:
            # We should use the logging module somehow and get the
            # logging info to the main process. Only print out the
            # sync time if we cannot make 100fps.
            log.warning('kaa.candy warning: sync took %0.4f sec' % sync_time)
        event.set()
        return False

# global mainloop object
mainloop = Mainloop()

class Server(object):
    """
    Backend server class
    """
    def __init__(self, name):
        self.ipc = kaa.rpc.Server(name)
        self.ipc.signals['client-connected'].connect(self.ipc_connected)
        self.ipc.register(self)
        self.widgets = {}
        self.initialized = False

    def ipc_connected(self, client):
        """
        Callback when the main app connects to the backend
        """
        client.signals['closed'].connect_once(self.ipc_disconnected)
        self.ipc = client

    def ipc_disconnected(self):
        """
        Callback when the main app disconnects from the backend
        """
        log.info('client disconnected, shutdown candy subprocess')
        sys.exit(0)

    @kaa.threaded(kaa.MAINTHREAD)
    def send_event(self, event, *args):
        self.ipc.rpc('event_%s' % event.replace('-', '_'), *args)

    @kaa.rpc.expose()
    def sync(self, tasks):
        """
        Sync callback from the candy application
        """
        if not self.initialized:
            self.initialized = True
            self.sync([
                    ('import', ('candy', os.path.dirname(__file__) + '/widgets')),
                    ('import', ('player', os.path.dirname(__file__) + '/player')),
                    ('import', ('stage', os.path.dirname(__file__) + '/stage')),
            ])
        queue = []
        for cmd, args in tasks:
            if cmd == 'freeze':
                # freeze meta-command after that the sync function in
                # the clutter thread is not allowed to go back to its
                # main loop to keep animations alive.
                queue.append(('freeze', None))
                continue
            func = getattr(self, 'cmd_' + cmd, None)
            if func:
                try:
                    delayed = func(*args)
                    if delayed:
                        queue.append(delayed)
                except Exception, e:
                    log.exception('sync error: %s%s', func, args)
            else:
                log.error('unsupported command: %s', cmd)
        if queue:
            # sync with the clutter thread
            event = threading.Event()
            # FIXME: idle_add does not work when animations are
            # running. It seems to be that clutter uses as much CPU
            # time as possible, nothing is idle.
            if kaa.get_thread_pool(kaa.GOBJECT) and kaa.get_thread_pool(kaa.GOBJECT).thread and \
                    not kaa.get_thread_pool(kaa.GOBJECT).thread.is_alive():
                log.error('gobject loop crashed')
                sys.exit(1)
            gobject.timeout_add(0, mainloop.sync, queue, event)
            event.wait()

    def cmd_import(self, name, path):
        """
        command for sync: import the module with path as name
        """
        return mainloop.imp, (name, path, self)

    def cmd_add(self, cls, wid):
        """
        command for sync: add a new widget based on cls with the given wid
        """
        self.widgets[wid] = eval(cls)()
        self.widgets[wid].server = self
        self.widgets[wid].wid = wid
        return self.widgets[wid].create, ()

    def cmd_call(self, wid, cmd, args):
        """
        command for sync: call the command with args on the given widget
        """
        args = list(args)
        for pos, value in enumerate(args):
            if isinstance(value, (str, unicode)) and value.startswith('candy:widget'):
                args[pos] = self.widgets[int(value[13:])]
        return getattr(self.widgets[wid], cmd), args

    def cmd_reparent(self, child, parent, pos):
        """
        command for sync: reparent the child
        """
        if parent:
            return self.widgets[child].reparent, (self.widgets[parent], self.widgets.get(pos))
        return self.widgets[child].reparent, (None, None)

    def cmd_position(self, wid, x, y):
        """
        command for sync: set a new position
        """
        w = self.widgets[wid]
        w.x = x
        w.y = y
        return w.set_position, ()

    def cmd_update(self, wid, modified):
        """
        command for sync: update the given widget
        """
        w = self.widgets[wid]
        for a, v in modified.items():
            setattr(w, a, v)
        w.prepare(modified)
        return w.update, (modified,)

    def cmd_delete(self, wid):
        """
        command for sync: delete the given widget
        """
        return self.widgets.pop(wid).delete, ()

    def cmd_active(self, state):
        """
        command for sync: set state (in)active
        """
        return self.widgets[-1].set_active, (state,)


#
# Set up logging module
#
logger = logging.getLogger()

# remove handler, we want to set the look and avoid duplicate handlers
for l in logger.handlers[:]:
    logger.removeHandler(l)

# set stdout logging
formatter = logging.Formatter('%(levelname)s candy:%(module)s(%(lineno)s): %(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)

if sys.argv[2]:
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s [%(name)6s] %(filename)s %(lineno)s: %(message)s')
    handler = logging.handlers.RotatingFileHandler(sys.argv[2], maxBytes=1000000, backupCount=2)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

# set log level
logger.setLevel(logging.INFO)

try:
    kaa.main.init('generic')
    kaa.gobject_set_threaded(mainloop)
    # start server and keep its reference to avoid it being deleted by
    # GC before the actual connect.
    s = Server(sys.argv[1])
    kaa.main.run()
except Exception, e:
    log.exception('backend shutdown')



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# progressbar.py - progressbar widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Progressbar' ]

# kaa imports
from kaa.utils import property

# kaa.candy imports
from .. import is_template
from rectangle import Rectangle
from group import AbstractGroup
from widget import Widget

class Progressbar(AbstractGroup):
    """
    Widget showing a progressbar. Only the bar is drawn, the border has
    to be created ouside this widget.
    """
    candyxml_name = 'progressbar'

    __max = 0
    __progress = 0

    def __init__(self, pos=None, size=None, progress=None, context=None):
        super(Progressbar, self).__init__(pos, size, context)
        if is_template(progress):
            progress = progress()
        if progress is None:
            # we have no progress bar, use a simple rectangle with its
            # default values. In most cases this is wrong
            progress = Rectangle()
        self.__bar = progress
        self.__bar.x = 0
        self.__bar.y = 0
        self.add(self.__bar)

    @property
    def max(self):
        return self.__max

    @max.setter
    def max(self, value):
        self.__max = value

    @property
    def progress(self):
        return self.__progress

    @progress.setter
    def progress(self, value):
        """
        Set a new progress and redraw the widget.
        """
        self.__progress = value
        self.queue_rendering()

    def inc(self):
        """
        Increase progress by one
        """
        self.__progress += 1
        self.queue_rendering()

    def sync_layout(self, size):
        """
        Sync layout changes and calculate intrinsic size based on the
        parent's size.
        """
        Widget.sync_layout(self, size)
        pos = float(self.__progress) / max(self.__max, self.__progress, 0.1)
        self.__bar.width = int(max(pos * self.width, 1))

    @classmethod
    def candyxml_parse(cls, element):
        """
        Parse the XML element for parameter to create the widget.
        """
        return super(Progressbar, cls).candyxml_parse(element).update(
            progress=element[0].xmlcreate())



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# mplayer.py - mplayer video widget
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# The widget does not use kaa.popcorn because gstreamer using
# kaa.candy and mplayer are too different to add support to
# kaa.popcorn without rewriting too much. Furthermore, kaa.popcorn
# seems to have a raise condition I cannot find and is too complex;
# much is not needed.
#
# Some parts of this code is copied from kaa.popcorn.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

# Python imports
import os
import sys
import re
import subprocess
import logging

# Clutter GI bindings
from gi.repository import Clutter as clutter, ClutterX11

# kaa.display for mplayer to draw to
import kaa.display

# kaa.candy backend code
import candy

# get logging object
log = logging.getLogger('candy')

# mplayer progress status line parser
RE_STATUS = re.compile(r'(?:V:\s*([\d.]+)|A:\s*([\d.]+)\s\W)(?:.*\s([\d.]+x))?')

class ArgumentList(list):
    """
    Argument list.
    """
    def __init__(self, args=()):
        if isinstance(args, basestring):
            args = args.split(' ')
        list.__init__(self, args)

    def __getslice__(self, i, j):
        return ArgumentList(list.__getslice__(self, i, j))

    def extend(self, arg):
        try:
            list.extend(self, arg.split(' '))
        except AttributeError:
            list.extend(self, arg)


SEEK_RELATIVE = 'SEEK_RELATIVE'
SEEK_ABSOLUTE = 'SEEK_ABSOLUTE'
SEEK_PERCENTAGE = 'SEEK_PERCENTAGE'

class Player(candy.Widget):
    """
    Mplayer video widget.
    """

    child = None

    def create(self):
        """
        Create the clutter object
        """
        self.cmd = ArgumentList('mplayer')
        self.obj = ClutterX11.TexturePixmap()
        # self.obj.hide()

    def update(self, modified):
        """
        Render the widget
        """
        pass

    #
    # control callbacks from the main process
    #

    @kaa.threaded(kaa.MAINTHREAD)
    def do_play(self):
        """
        Start playback
        """
        try:
            self.window = kaa.display.X11Window(size=(self.width,self.height))
            self.window.draw_rectangle((0,0), (self.width,self.height), '#000000')
            # self.window.set_fullscreen()
            self.window.set_cursor_visible(False)
            self.window.signals['key_press_event'].connect_weak(self.event_key)
            self.window.is_hidden = True
            self.streaminfo = {
                'audio': {},
                'subtitle': {},
                'is_menu': False,
                'sync': True
            }
            cmd = self.cmd[:]
            cmd.extend('-v -osdlevel 0 -slave -nolirc -nojoystick -identify -framedrop')
            # video playback
            cmd.extend('-fs -wid %s' % str(self.window.id))
            # vpdau
            if self.audio_only:
                cmd.extend('-vo null')
            else:
                if self.config['mplayer.vdpau']:
                    # Add deinterlacer. We disable it later when the
                    # video starts
                    cmd.extend('-vo vdpau:deint=2,xv,x11 -vc ffvc1vdpau,ffh264vdpau,')
                else:
                    # Add deinterlacer. We disable it later when the
                    # video starts
                    cmd.extend('-vf yadif')
            # passthrough
            if self.config['mplayer.passthrough']:
                cmd.extend('-ac hwac3,hwdts,')
            if self.uri.startswith('dvd://'):
                self.child = kaa.Process(cmd + [ '-nocache', '-dvd-device', self.uri[6:-1], 'dvdnav://' ])
            else:
                self.child = kaa.Process(cmd + [ self.uri ])
            self.child.delimiter = ['\r', '\n']
            self.child.stop_command = 'quit\nquit\n'
            self.child.signals['exited'].connect_weak_once(self.event_exit)
            self.child.signals['readline'].connect_weak(self.event_stdout)
            # lower the window before showing it. If kaa.candy is also
            # fullscreen the window is hidden until the playback
            # starts
            self.window.lower()
            self.window.show()
            self.window.lower()
            self.child.start()
        except Exception, e:
            # We should use the logging module somehow and get the
            # logging info to the main process.
            log.exception('unable to start mplayer')

    @kaa.threaded(kaa.MAINTHREAD)
    def do_stop(self):
        """
        Stop playback
        """
        if not self.window:
            return
        self.child.write('quit\n')

    @kaa.threaded(kaa.MAINTHREAD)
    def do_pause(self):
        """
        Pause playback
        """
        self.child.write('pause\n')

    @kaa.threaded(kaa.MAINTHREAD)
    def do_resume(self):
        """
        Resume playback
        """
        self.child.write('pause\n')

    def do_seek(self, value, type):
        """
        Seek to the given position
        """
        if self.child:
            s = [SEEK_RELATIVE, SEEK_PERCENTAGE, SEEK_ABSOLUTE]
            return self.child.write('seek %s %s\n' % (value, s.index(type)))
        # Playback is not started. We can use the seek command line
        # argument.
        if type in (SEEK_RELATIVE, SEEK_ABSOLUTE):
            # sorry, we do not support more here
            self.cmd.extend('-ss %s' % value)

    def do_set_audio(self, idx):
        """
        Set the audio stream
        """
        if self.child:
            return self.child.write('switch_audio %s\n' % idx)

    def do_set_subtitle(self, idx):
        """
        Set the subtitle stream (-1 == none)
        """
        if self.child:
            return self.child.write('sub_select %s\n' % idx)

    def do_set_aspect(self, aspect):
        """
        Set the aspect ratio
        """
        pass

    def do_set_deinterlace(self, value):
        """
        Turn on/off deinterlacing
        """
        if self.child:
            return self.child.write('set_property deinterlace %d\n' % int(value))

    def do_nav_command(self, cmd):
        """
        Send DVD navigation command
        """
        self.child.write('dvdnav %s\n' % cmd)

    #
    # events from mplayer
    #

    def event_key(self, key):
        """
        Key press in the player window
        """
        self.server.send_event('key_press', key)

    @kaa.threaded(kaa.GOBJECT)
    def hook(self):
        self.obj.set_window(self.window.id, True)
        self.obj.set_automatic(True)

    def event_stdout(self, line):
        """
        Stdout line from the running mplayer process
        """
        if line.startswith('V:') or line.startswith('A:'):
            m = RE_STATUS.search(line)
            if m:
                if self.window.is_hidden:
                    # We just started. Disable the deinterlacer and
                    # show the window
                    self.do_set_deinterlace(False)
                    # self.window.raise_()
                    self.window.draw_rectangle((0,0), (self.width,self.height), '#000000')
                    self.hook()
                    self.window.is_hidden = False
                if self.streaminfo['sync']:
                    self.send_widget_event('streaminfo', self.streaminfo)
                    self.streaminfo['sync'] = False
            self.send_widget_event('progress', float(m.groups()[0]))
            return
        if line.startswith('ID_') or line.startswith('DVDNAV'):
            self.streaminfo['sync'] = True
            line = line.strip()
            if line.startswith('ID_AUDIO_ID'):
                aid = int(line.strip().split('=')[1])
                if not aid in self.streaminfo['audio']:
                    self.streaminfo['audio'][aid] = None
            if line.startswith('ID_AID') and line.find('_LANG=') > 0:
                aid, lang = line.split('=')
                self.streaminfo['audio'][int(aid[7:-5])] = lang
            if line.startswith('ID_SUBTITLE_ID'):
                sid = int(line.strip().split('=')[1])
                if not sid in self.streaminfo['subtitle']:
                    self.streaminfo['subtitle'][sid] = None
            if line.startswith('ID_SID') and line.find('_LANG=') > 0:
                sid, lang = line.split('=')
                self.streaminfo['subtitle'][int(sid[7:-5])] = lang
            if line.startswith('DVDNAV_TITLE_IS_MENU'):
                self.streaminfo['is_menu'] = True
            if line.startswith('DVDNAV_TITLE_IS_MOVIE'):
                self.streaminfo['is_menu'] = False

    def event_exit(self, code):
        """
        The player has stopped
        """
        if not self.window:
            return
        self.child.signals['readline'].disconnect(self.event_stdout)
        self.window.signals['key_press_event'].disconnect(self.event_key)
        self.window.hide()
        self.server.send_event('widget_call', self.wid, 'finished')
        self.window = None
        self.child = None



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# rectangle.py - rectangle widget
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# FIXME: instead of using cairo cogl may be faster. But calling
# get_cogl_texture from a Texture does not work for some strange
# reason. This should be fixed here in the future.
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2012-2013 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Rectangle' ]

# kaa.candy imports
import image

class Rectangle(image.CairoTexture):

    def draw(self, cr):
        """
        Render the cairo context
        """
        stroke = self.border_size or 1
        width = self.width - 2 * stroke
        height = self.height - 2 * stroke
        radius = min(self.radius, width, height)
        x0 = stroke
        y0 = stroke
        x1 = int(x0 + width)
        y1 = int(y0 + height)
        if self.color:
            cr.set_source_rgba(*self.color.to_cairo())
            cr.set_line_width(stroke)
            cr.move_to  (x0, y0 + radius)
            cr.curve_to (x0, y0, x0 , y0, x0 + radius, y0)
            cr.line_to (x1 - radius, y0)
            cr.curve_to (x1, y0, x1, y0, x1, y0 + radius)
            cr.line_to (x1 , y1 - radius)
            cr.curve_to (x1, y1, x1, y1, x1 - radius, y1)
            cr.line_to (x0 + radius, y1)
            cr.curve_to (x0, y1, x0, y1, x0, y1- radius)
            cr.close_path()
            cr.fill()
        if self.border_size and self.border_color:
            cr.set_source_rgba(*self.border_color.to_cairo())
            cr.set_line_width(stroke)
            cr.move_to  (x0, y0 + radius)
            cr.curve_to (x0 , y0, x0 , y0, x0 + radius, y0)
            cr.line_to (x1 - radius, y0)
            cr.curve_to (x1, y0, x1, y0, x1, y0 + radius)
            cr.line_to (x1 , y1 - radius)
            cr.curve_to (x1, y1, x1, y1, x1 - radius, y1)
            cr.line_to (x0 + radius, y1)
            cr.curve_to (x0, y1, x0, y1, x0, y1- radius)
            cr.close_path()
            cr.stroke()
        return True



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# stage.py - Stage class as base for all widgets
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2011-2013 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

# python imports
import os
import time
import sys
import fcntl
import subprocess
import threading
import logging

# kaa imports
import kaa
import kaa.rpc

# kaa.candy imports
from widgets import Group, Widget, POSSIBLE_PLAYER

import candyxml

# get logging object
log = logging.getLogger('kaa.candy')

# turn on internal debug
DEBUG = False

class Layer(Group):
    """
    Group on the stage as parent for widgets added to the stage.
    """

    STATUS_NEW = 'STATUS_NEW'
    STATUS_ACTIVE = 'STATUS_ACTIVE'
    STATUS_DESTROYED = 'STATUS_DESTROYED'

    _state = STATUS_NEW

    def __init__(self, size, widgets=[], context=None):
        super(Layer, self).__init__(size=size, widgets=widgets, context=context)

    def __reset__(self):
        """
        Internal function when the candy backend becomes invalid and
        needs to be restarted.
        """
        super(Layer, self).__reset__()
        self._state = Layer.STATUS_NEW
        Widget._candy_sync_reparent.remove(self)

    @property
    def parent(self):
        return self.stage


class Stage(kaa.Object):
    """
    The stage class is the visible window
    """

    BACKEND_DOWN = 'BACKEND_DOWN'
    BACKEND_INITIALIZING = 'BACKEND_INITIALIZING'
    BACKEND_RUNNING = 'BACKEND_RUNNING'

    __kaasignals__ = {
        'key-press':
            '''
            Emitted when a key in the window is pressed
            '''
    }

    active = True

    def __init__(self, size, name, logfile=''):
        super(Stage, self).__init__()
        self.name = 'candy-backend-%s' % name
        self.logfile = logfile
        self.size = size
        self.scale = None
        self.backend_state = Stage.BACKEND_DOWN
        self.backend_time = 0
        # import cache for restart
        self._candy_restart_import = []
        # create the base widget
        self.layer = [ Layer(size=size) ]
        # We need the render pipe, the 'step' signal is not enough. It
        # is not triggered between timer and select and a change done
        # in a timer may get lost.
        self._render_pipe = os.pipe()
        fcntl.fcntl(self._render_pipe[0], fcntl.F_SETFL, os.O_NONBLOCK)
        fcntl.fcntl(self._render_pipe[1], fcntl.F_SETFL, os.O_NONBLOCK)
        kaa.IOMonitor(self.__sync).register(self._render_pipe[0])
        os.write(self._render_pipe[1], '1')
        self._start_backend()

    @kaa.coroutine()
    def _start_backend(self):
        # spawn the backend process
        args = [ 'python', os.path.dirname(__file__) + '/backend/main.py', self.name, self.logfile ]
        self._candy_dirty = True
        self.server = subprocess.Popen(args, stdout=sys.stdout, stderr=sys.stderr)
        retry = 50
        if os.path.exists(kaa.tempfile(self.name)):
            os.unlink(kaa.tempfile(self.name))
        while True:
            try:
                self.ipc = kaa.rpc.connect(self.name)
                yield kaa.inprogress(self.ipc)
                break
            except Exception, e:
                retry -= 1
                if retry == 0:
                    raise e
                time.sleep(0.1)
        self.ipc.signals['closed'].connect_weak_once(self._ipc_disconnect)
        self.ipc.register(self)
        self.backend_state = Stage.BACKEND_INITIALIZING
        self.commands = []
        self.tasks = []
        os.write(self._render_pipe[1], '1')

    def __reset__(self):
        """
        Reset the stage during disconnect
        """
        pass

    @kaa.coroutine()
    def _ipc_disconnect(self):
        """
        Callback when the backend disconnects (crash). This means we
        need to restart it.
        """
        if time.time() - self.backend_time < 5:
            log.error('two backend errors in 5 seconds, giving up')
            sys.exit(1)
        log.error('backend error, restarting')
        self.backend_state = Stage.BACKEND_DOWN
        self.backend_time = time.time()
        self.ipc = None
        self.__reset__()
        # remove dead layer
        for layer in self.layer[:]:
            if layer._state == Layer.STATUS_DESTROYED:
                self.layer.remove(layer)
        # start the backend again
        yield self._start_backend()
        # reset widget status
        Widget._candy_sync_new = []
        Widget._candy_sync_reparent = []
        Widget._candy_sync_delete = []
        for widget in Widget._candy_all_widgets:
            widget.__reset__()
        self.__sync()

    def add_layer(self, layer=None, sibling=None):
        """
        Add a new layer and return its id
        """
        if layer is None:
            layer = Layer(size=self.size)
        if self.scale:
            (layer.scale_x, layer.scale_y), (layer.width, layer.height) = self.scale
        if sibling:
            self.layer.insert(self.layer.index(sibling)+1, layer)
            layer.above_sibling(sibling)
        else:
            self.layer.append(layer)
        return len(self.layer) - 1

    def destroy_layer(self, layer):
        """
        Destroy the given layer
        """
        if isinstance(layer, (int, long)):
            layer = self.layer[layer]
        layer._state = Layer.STATUS_DESTROYED

    def hide(self):
        """
        Hide the clutter actors to make it possibile to use the X11
        window for other applications such as a video player.
        """
        for layer in self.layer:
            self.queue_command(layer._candy_id, 'hide', ())

    def show(self):
        """
        Show the clutter actors again after calling hide.
        """
        for layer in self.layer:
            self.queue_command(layer._candy_id, 'show', ())
        self.queue_command(-1, 'ensure_redraw', ())

    def add(self, *widgets, **kwargs):
        """
        Add the given widgets to the stage
        """
        self.layer[kwargs.get('layer', 0)].add(*widgets)

    def remove(self, *widgets, **kwargs):
        """
        Remove the given widgets from the stage
        """
        self.layer[kwargs.get('layer', 0)].remove(*widgets)

    def get_widget(self, name):
        """
        Get child element with the given name. For group children this
        function will search recursive.

        @param name: name of the child
        @returns: widget or None
        """
        for layer in self.layer:
            widget = layer.get_widget(name)
            if widget:
                return widget

    def set_active(self, state):
        """
        Set the Stage active or inactive. If inactive the stage is
        faded out and keys are not handled anymore. Usefull when the
        application is busy somehow.
        """
        if state != self.active:
            self.ipc.rpc('sync', [('active', (state, ))])
            self.active = state
            if state and self.tasks:
                # some queued tasks, rerun sync
                os.write(self._render_pipe[1], '1')

    def queue_rendering(self):
        """
        Queue sync for rendering
        """
        if not self._candy_dirty:
            self._candy_dirty = True
            os.write(self._render_pipe[1], '1')

    def queue_command(self, candy_id, cmd, args):
        """
        Queue sync for a command
        """
        self.commands.append((candy_id, cmd, args))
        self.queue_rendering()

    def __sync(self):
        """
        Sync callback when something changed
        """
        # read the socket to handle the sync
        try:
            os.read(self._render_pipe[0], 1)
        except OSError:
            pass
        if self.backend_state == Stage.BACKEND_DOWN:
            return
        self._candy_dirty = False
        tasks = []
        # check for new imports
        if self.backend_state == Stage.BACKEND_INITIALIZING:
            for i in self._candy_restart_import:
                tasks.append(('import', i))
        while Widget._candy_import:
            module = Widget._candy_import.pop()
            self._candy_restart_import.append(module)
            tasks.append(('import', module))
        if tasks:
            # make sure everything is imported before we need it
            self.ipc.rpc('sync', tasks)
            tasks = []
        # prepare all widgets for the sync
        for layer in self.layer:
            layer.sync_prepare()
        # Create new widgets. The create functions at the backend only
        # create the widgets and not fill it with the actual
        # content. This is done later. Remember the ids of the widget
        # re-arrange the calls later.
        new_widgets = []
        while Widget._candy_sync_new:
            widget = Widget._candy_sync_new.pop(0)
            widget.stage = self
            widget.backend.stage = self
            while widget.backend.queue:
                self.commands.append(widget.backend.queue.pop(0))
            tasks.append(('add', (widget.candy_backend, widget._candy_id)))
            new_widgets.append(widget._candy_id)
        # create stage object if needed
        if self.backend_state == Stage.BACKEND_INITIALIZING:
            self.backend_state = Stage.BACKEND_RUNNING
            tasks.append(('add', ('stage.Stage', -1)))
            tasks.append(('call', (-1, 'init', (self.size, ))))
        # Change parents for the widgets. Remember the needed calls in
        # a seperate variable. We need to reparent before updaing
        # everything here, but maybe do it later on the backend.
        tasks_reparent = []
        while Widget._candy_sync_reparent:
            widget = Widget._candy_sync_reparent.pop(0)
            if widget.parent:
                tasks_reparent.append(('reparent', (widget._candy_id, widget.parent._candy_id, widget._candy_stack)))
            else:
                tasks_reparent.append(('reparent', (widget._candy_id, None, None)))
        # sync all children
        tasks_update = []
        for layer in self.layer[:]:
            if layer._state == Layer.STATUS_NEW:
                tasks.append(('reparent', (layer._candy_id, -1, None)))
                layer._state = Layer.STATUS_ACTIVE
            if layer._state == Layer.STATUS_DESTROYED:
                tasks.append(('reparent', (layer._candy_id, None, None)))
                self.layer.remove(layer)
            else:
                layer.__sync__(tasks_update)
        # Now the tricky part. All create, update and move functions
        # are called in the clutter thread at the backend. If it takes
        # too long, running animations may look strange. But for new
        # objects without parents on the backend we can go back to the
        # clutter main loop from time to time and the user won't see a
        # partial update. Therefore search for new widgets here that
        # gets updated and add the update calls to the tasks list
        # right after the creation. After that add a fake "freeze"
        # call after which going back to the main loop is not allowed
        # anymore. After that freeze add the reparent tasks and all
        # other tasks affecting visible actors.
        for t in tasks_update[:]:
            if t[0] == 'update' and t[1][0] in new_widgets:
                tasks.append(t)
                tasks_update.remove(t)
        self.tasks = self.tasks + tasks + [('freeze', None)] + tasks_reparent + tasks_update
        # add commands for the widgets
        while self.commands:
            self.tasks.append(('call', self.commands.pop(0)))
        while Widget._candy_sync_delete:
            self.tasks.append(('delete', (Widget._candy_sync_delete.pop(0),)))
        # now send everything to the backend
        if self.tasks and self.active:
            if DEBUG:
                print 'sync'
                for t in self.tasks:
                    print '', t
                print
            try:
                if self.ipc:
                    self.ipc.rpc('sync', self.tasks)
                self.tasks = []
            except kaa.rpc.NotConnectedError, e:
                self._ipc_disconnect()

    def set_content_geometry(self, size):
        """
        Set the geometry. This will scale the group in the stage
        """
        if isinstance(size, (str, unicode)):
            size = int(size.split('x')[0]), int(size.split('x')[1])
        self.queue_rendering()
        self.scale = (float(self.size[0]) / size[0], float(self.size[1]) / size[1]), size
        for layer in self.layer:
            (layer.scale_x, layer.scale_y), (layer.width, layer.height) = self.scale

    @kaa.rpc.expose()
    def event_key_press(self, key):
        """
        Callback from the backend on key press
        """
        if self.active:
            self.signals['key-press'].emit(key)

    @kaa.rpc.expose()
    def event_widget_call(self, wid, func, *args, **kwargs):
        """
        Callback for a widget
        """
        for layer in self.layer:
            widget = layer.get_widget_by_id(wid)
            if widget:
                getattr(widget, 'event_%s' % func)(*args, **kwargs)

    @kaa.rpc.expose()
    def event_player(self, player):
        """
        Callback on init to list all possible player
        """
        for p in player:
            POSSIBLE_PLAYER.append(p)

    def candyxml(self, data):
        """
        Load a candyxml file based on the given screen resolution.

        @param data: filename of the XML file to parse or XML data
        @returns: root element attributes and dict of parsed elements
        """
        attr, templates = candyxml.parse(data)
        if 'geometry' in attr:
            self.set_content_geometry(attr['geometry'])
        return attr, templates



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# vieo.py - video widget
# -----------------------------------------------------------------------------
# This file is imported by the backend process in the clutter
# mainloop. Importing and using clutter is thread-safe.
#
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2012 Dirk Meyer
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'player', 'Video', 'Audio' ]

import os
from gi.repository import Clutter as clutter

import kaa.metadata
import group

player = {}

class Video(group.Group):

    player_widget = None
    unref_obj = None

    attributes = [ 'width', 'height', 'uri', 'config', 'audio_only', 'player' ]

    def create(self):
        super(Video, self).create()
        self.delayed = []

    def prepare(self, modified):
        """
        Prepare rendering
        """
        if 'player' in modified and self.player:
            if self.player_widget:
                self.unref_obj = self.player_widget.obj
            self.player_widget = player[self.player]()
            self.player_widget.x = self.player_widget.y = 0
            self.player_widget.wid = self.wid
            self.player_widget.server = self.server
        for a in self.attributes:
            setattr(self.player_widget, a, getattr(self, a))
        self.player_widget.prepare(modified)

    def update(self, modified):
        """
        Render the widget
        """
        super(Video, self).update(modified)
        if self.unref_obj:
            self.obj.remove_actor(self.unref_obj)
            self.unref_obj = None
        if 'width' in modified or 'height' in modified:
            self.obj.set_clip(0, 0, self.width, self.height)
        if 'uri' in modified and self.uri:
            self.streaminfo = {
                'audio': {},
                'subtitle': {},
                'is_menu': False,
                'sync': True
            }
            metadata = kaa.metadata.parse(self.uri)
            if metadata and 'audio' in metadata:
                # video item, not audio only
                for audio in metadata.audio:
                    self.streaminfo['audio'][audio.id] = None if audio.langcode == 'und' else audio.langcode
                for sub in metadata.subtitles:
                    self.streaminfo['subtitle'][sub.id] = None if sub.langcode == 'und' else sub.langcode
            self.send_widget_event('streaminfo', self.streaminfo)
            if self.config and self.config.get('refresh-rate', None):
                os.system('xrandr -r %s' % self.config['refresh-rate'])
        if 'player' in modified and self.player:
            self.player_widget.create()
            self.obj.add_actor(self.player_widget.obj)
        if self.player_widget:
            self.player_widget.update(modified)

    #
    # control callbacks from the main process
    #

    def __getattr__(self, attr):
        return getattr(self.player_widget, attr)


class Audio(Video):

    attributes = Video.attributes + [ 'visualisation' ]



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# clone.py - clone widget for another widget
# -----------------------------------------------------------------------------
# kaa-candy - Fourth generation Canvas System using Clutter as backend
# Copyright (C) 2013 Dirk Meyer
#
# First Version: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Based on various previous attempts to create a canvas system for
# Freevo by Dirk Meyer and Jason Tackaberry.  Please see the file
# AUTHORS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Clone' ]

# python imports
import logging

# kaa imports
import kaa

# kaa.candy imports
from widget import Widget

# get logging object
log = logging.getLogger('kaa.candy')

class Clone(Widget):
    """
    """
    candy_backend = 'candy.Clone'

    def __init__(self, pos=None, master=None):
        super(Clone, self).__init__(pos, None, context={})
        self.master = master
        self.__candy_set_master = True

    def sync_layout(self, size):
        """
        Sync layout is based on the master object
        """
        super(Clone, self).sync_layout(size)
        self.intrinsic_size = self.master.intrinsic_size

    def __sync__(self, tasks):
        """
        Internal function to add the changes to the list of tasks for
        the backend.
        """
        super(Clone, self).__sync__(tasks)
        if self.__candy_set_master:
            self.backend.call('set_master', 'candy:widget:%s' % self.master._candy_id)
            self.__candy_set_master = False



# Usage: test/beacon.py path-to-imagedir
# Note: kaa.beacon daemon must be running
import sys
import os

import kaa
import kaa.candy
import feedparser
import kaa.beacon

# we use the following xml (file) as gui. There is one widget:
# a group called "thumbnails" with a label and a grid in it.
# The grid needs a template for each cell which is again a
# group with a label and an image.
xml = '''
<candyxml geometry="800x600">
    <group name="thumbnails" x="10" y="10" width="780" height="580">
        <label font="Vera:24" color="0xcccccc">
            <properties xalign="center"/>
            $title
        </label>
        <grid style="selection" y="50" height="530"
            cell-width="120" cell-height="80" items="items" cell-item="item"
            orientation="vertical">
            <properties name="items"/>
            <thumbnail thumbnail="item">
                <properties xalign="center" yalign="center"/>
            </thumbnail>
            <selection>
                <rectangle color="0x6666cc" width="126" height="86"/>
            </selection>
        </grid>
    </group>
</candyxml>
'''

# create a stage window and parse the xml file
stage = kaa.candy.Stage((800,600), 'beacon')
candy = stage.candyxml(xml)[1]

@kaa.coroutine()
def main():
    query = yield (yield kaa.beacon.get(sys.argv[1])).list()

    # this is the context for the images widget
    context = kaa.candy.Context(title=os.path.basename(sys.argv[1]), items=query)
    group = candy.group.thumbnails(context=context)
    stage.add(group)
    grid = group.get_widget('items')

    # if 1:
    #     # add effects and hide selection rectangle
    #     grid.behave('opacity', 80, 255).behave('scale', (1, 1), (1.1, 1.1))
    #     grid.selection.opacity = 0
    # if 0:
    #     # add effects and hide selection rectangle
    #     grid.behave('scale', (1, 1), (1.5, 1.5))

    yield kaa.delay(0.5)

    # now we move the selection
    print 'move selection to the right'
    grid.select((2, 0), 1)
    yield kaa.delay(1.5)

    print 'move selection down'
    grid.select((2, 2), 1)
    yield kaa.delay(1.5)

    print 'move selection down and scroll at the same time to make it look'
    print 'like the selection is standing still.'
    grid.select((2, 4), 1)
    grid.scroll_by((0, 2), 1)
    yield kaa.delay(1.5)
    print 'move selection down fast'
    grid.select((2, 5), 0.3)
    yield kaa.delay(1.5)
    print 'move selection and start scrolling with a different speed'
    grid.select((5, 5), 1)
    yield kaa.delay(0.3)
    grid.scroll_by((4, 0), 4)
    yield kaa.delay(2)
    grid.select((5, 3), 1)
    yield kaa.delay(2)
    print 'move selection very slowly'
    grid.select((7, 3), 3)
    yield kaa.delay(3.5)
    print 'scroll to get selection to the left side'
    grid.scroll_by((3, 0), 1)
    yield kaa.delay(2)
    print 'move selection fast to the right'
    grid.select((12, 3), 0.8)
    yield kaa.delay(1)
    print 'and back'
    grid.select((7, 3), 1)
    yield kaa.delay(0.3)
    print 'and revert while moving'
    grid.select((12, 3), 1)
    yield kaa.delay(1)
    print 'and back again very fast'
    grid.select((7, 3), 0.3)
    yield kaa.delay(0.5)
    print 'left again and start scrolling'
    grid.select((16, 3), 0.8)
    yield kaa.delay(0.3)
    grid.scroll_by((4, 0), 0.5)
    yield kaa.delay(1.0)
    print 'change col and row'
    grid.select((11, 6), 1)

main()

# run the kaa mainloop, it takes some time to load all the images.
kaa.main.run()
